from abc import ABC, abstractmethod
import logging
import os
import tempfile

import pandas as pd
import tensorflow as tf

from .. import normalisation
from ..vector_model import VectorRegressionModel

log = logging.getLogger(__name__)


class TensorFlowSession:
    session = None
    _isKerasSessionSet = False

    @classmethod
    def configure_session(cls, gpu_allow_growth=True, gpu_per_process_memory_fraction=None):
        tf_config = tf.compat.v1.ConfigProto()
        tf_config.gpu_options.allow_growth = gpu_allow_growth  # dynamically grow the memory used on the GPU
        tf_config.log_device_placement = False
        if gpu_per_process_memory_fraction is not None:
            tf_config.gpu_options.per_process_gpu_memory_fraction = gpu_per_process_memory_fraction  # in case we get CUDNN_STATUS_INTERNAL_ERROR
        cls.session = tf.compat.v1.Session(config=tf_config)

    @classmethod
    def set_keras_session(cls, allow_default=True):
        """
        Sets the (previously configured) session for use with keras if it has not been previously been set.
        If no session has been configured, the parameter allowDefault controls whether it is admissible to create a session with default
        parameters.

        :param allow_default: whether to configure, for the case where no session was previously configured, a new session with the defaults.
        """
        if cls.session is None:
            if allow_default:
                log.info("No TensorFlow session was configured. Creating a new session with default values.")
                cls.configure_session()
            else:
                raise Exception(f"The session has not yet been configured. Call {cls.__name__}.{cls.configure_session.__name__} beforehand")
        if not cls._isKerasSessionSet:
            tf.keras.backend.set_session(cls.session)
            cls._isKerasSessionSet = True


class KerasVectorRegressionModel(VectorRegressionModel, ABC):
    """An abstract simple model which maps vectors to vectors and works on pandas.DataFrames (for inputs and outputs)"""

    def __init__(self, normalisation_mode: normalisation.NormalisationMode, loss, metrics, optimiser,
            batch_size=64, epochs=1000, validation_fraction=0.2):
        """
        :param normalisation_mode:
        :param loss:
        :param metrics:
        :param optimiser:
        :param batch_size:
        :param epochs:
        :param validation_fraction:
        """
        super().__init__()
        self.normalisation_mode = normalisation_mode
        self.batch_size = batch_size
        self.epochs = epochs
        self.optimiser = optimiser
        self.loss = loss
        self.metrics = list(metrics)
        self.validation_fraction = validation_fraction

        self.model = None
        self.input_scaler = None
        self.output_scaler = None
        self.training_history = None

    def __str__(self):
        params = dict(normalisationMode=self.normalisation_mode, optimiser=self.optimiser, loss=self.loss, metrics=self.metrics,
            epochs=self.epochs, validationFraction=self.validation_fraction, batchSize=self.batch_size)
        return f"{self.__class__.__name__}{params}"

    @abstractmethod
    def _create_model(self, input_dim, output_dim):
        """
        Creates a keras model

        :param input_dim: the number of input dimensions
        :param output_dim: the number of output dimensions
        :return: the model
        """
        pass

    def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        # normalise data
        self.input_scaler = normalisation.VectorDataScaler(inputs, self.normalisation_mode)
        self.output_scaler = normalisation.VectorDataScaler(outputs, self.normalisation_mode)
        norm_inputs = self.input_scaler.get_normalised_array(inputs)
        norm_outputs = self.output_scaler.get_normalised_array(outputs)

        # split data into training and validation set
        train_split = int(norm_inputs.shape[0] * (1-self.validation_fraction))
        train_inputs = norm_inputs[:train_split]
        train_outputs = norm_outputs[:train_split]
        val_inputs = norm_inputs[train_split:]
        val_outputs = norm_outputs[train_split:]

        # create and fit model
        TensorFlowSession.set_keras_session()
        model = self._create_model(inputs.shape[1], outputs.shape[1])
        model.compile(optimizer=self.optimiser, loss=self.loss, metrics=self.metrics)
        temp_file_handle, temp_file_path = tempfile.mkstemp(".keras.model")
        try:
            os.close(temp_file_handle)
            checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(temp_file_path, monitor='val_loss', save_best_only=True,
                save_weights_only=True)
            self.training_history = model.fit(train_inputs, train_outputs, batch_size=self.batch_size, epochs=self.epochs, verbose=2,
                validation_data=(val_inputs, val_outputs), callbacks=[checkpoint_callback])
            model.load_weights(temp_file_path)
        finally:
            os.unlink(temp_file_path)
        self.model = model

    def _predict(self, inputs: pd.DataFrame) -> pd.DataFrame:
        x = self.input_scaler.get_normalised_array(inputs)
        y = self.model.predict(x)
        y = self.output_scaler.get_denormalised_array(y)
        return pd.DataFrame(y, columns=self.output_scaler.dimension_names)



import warnings
import logging


log = logging.getLogger(__name__)


def deprecated(message):
    """
    This is a decorator for functions to mark them as deprecated, issuing a warning when the function is called

    :param message: message, which can indicate the new recommended approach or reason for deprecation
    :return: decorated function
    """
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            msg = "{} is a deprecated function. {}".format(func.__name__, message)
            if logging.Logger.root.hasHandlers():
                log.warning(msg)
            else:
                warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
                warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)
        return deprecated_func
    return deprecated_decorator



import logging
from abc import ABC, abstractmethod
from typing import Tuple, Sequence, TypeVar, Generic

import numpy as np
import pandas as pd
import scipy.stats
from sklearn.model_selection import StratifiedShuffleSplit

from .util.string import ToStringMixin

log = logging.getLogger(__name__)

T = TypeVar("T")


class BaseInputOutputData(Generic[T], ABC):
    def __init__(self, inputs: T, outputs: T):
        """
        :param inputs: expected to have shape and __len__
        :param outputs: expected to have shape and __len__
        """
        if len(inputs) != len(outputs):
            raise ValueError("Lengths do not match")
        self.inputs = inputs
        self.outputs = outputs

    def __len__(self):
        return len(self.inputs)

    @abstractmethod
    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        pass


class InputOutputArrays(BaseInputOutputData[np.ndarray]):
    def __init__(self, inputs: np.ndarray, outputs: np.ndarray):
        super().__init__(inputs, outputs)

    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs[indices]
        outputs = self.outputs[indices]
        return InputOutputArrays(inputs, outputs)

    def to_torch_data_loader(self, batch_size=64, shuffle=True):
        try:
            import torch
            from torch.utils.data import DataLoader, TensorDataset
        except ImportError:
            raise ImportError(f"Could not import torch, did you install it?")
        dataset = TensorDataset(torch.tensor(self.inputs), torch.tensor(self.outputs))
        return DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)


class InputOutputData(BaseInputOutputData[pd.DataFrame], ToStringMixin):
    """
    Holds input and output data for learning problems
    """
    def __init__(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        super().__init__(inputs, outputs)

    def _tostring_object_info(self) -> str:
        return f"N={len(self.inputs)}, numInputColumns={len(self.inputs.columns)}, numOutputColumns={len(self.outputs.columns)}"

    @classmethod
    def from_data_frame(cls, df: pd.DataFrame, *output_columns: str) -> "InputOutputData":
        """
        :param df: a data frame containing both input and output columns
        :param output_columns: the output column name(s)
        :return: an InputOutputData instance with inputs and outputs separated
        """
        inputs = df[[c for c in df.columns if c not in output_columns]]
        outputs = df[list(output_columns)]
        return cls(inputs, outputs)

    def filter_indices(self, indices: Sequence[int]) -> __qualname__:
        inputs = self.inputs.iloc[indices]
        outputs = self.outputs.iloc[indices]
        return InputOutputData(inputs, outputs)

    def filter_index(self, index_elements: Sequence[any]) -> __qualname__:
        inputs = self.inputs.loc[index_elements]
        outputs = self.outputs.loc[index_elements]
        return InputOutputData(inputs, outputs)

    @property
    def input_dim(self):
        return self.inputs.shape[1]

    @property
    def output_dim(self):
        return self.outputs.shape[1]

    def compute_input_output_correlation(self):
        correlations = {}
        for outputCol in self.outputs.columns:
            correlations[outputCol] = {}
            output_series = self.outputs[outputCol]
            for inputCol in self.inputs.columns:
                input_series = self.inputs[inputCol]
                pcc, pvalue = scipy.stats.pearsonr(input_series, output_series)
                correlations[outputCol][inputCol] = pcc
        return correlations


TInputOutputData = TypeVar("TInputOutputData", bound=BaseInputOutputData)


class DataSplitter(ABC, Generic[TInputOutputData]):
    @abstractmethod
    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        pass


class DataSplitterFractional(DataSplitter):
    def __init__(self, fractional_size_of_first_set: float, shuffle=True, random_seed=42):
        if not 0 <= fractional_size_of_first_set <= 1:
            raise Exception(f"invalid fraction: {fractional_size_of_first_set}")
        self.fractionalSizeOfFirstSet = fractional_size_of_first_set
        self.shuffle = shuffle
        self.randomSeed = random_seed

    def split_with_indices(self, data) -> Tuple[Tuple[Sequence[int], Sequence[int]], Tuple[TInputOutputData, TInputOutputData]]:
        num_data_points = len(data)
        split_index = int(num_data_points * self.fractionalSizeOfFirstSet)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(num_data_points)
        else:
            indices = range(num_data_points)
        indices_a = indices[:split_index]
        indices_b = indices[split_index:]
        a = data.filter_indices(list(indices_a))
        b = data.filter_indices(list(indices_b))
        return (indices_a, indices_b), (a, b)

    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        _, (a, b) = self.split_with_indices(data)
        return a, b


class DataSplitterFromDataFrameSplitter(DataSplitter[InputOutputData]):
    """
    Creates a DataSplitter from a DataFrameSplitter, which can be applied either to the input or the output data.
    It supports only InputOutputData, not other subclasses of BaseInputOutputData.
    """
    def __init__(self, data_frame_splitter: "DataFrameSplitter", fractional_size_of_first_set: float, apply_to_input=True):
        """
        :param data_frame_splitter: the splitter to apply
        :param fractional_size_of_first_set: the desired fractional size of the first set when applying the splitter
        :param apply_to_input: if True, apply the splitter to the input data frame; if False, apply it to the output data frame
        """
        self.dataFrameSplitter = data_frame_splitter
        self.fractionalSizeOfFirstSet = fractional_size_of_first_set
        self.applyToInput = apply_to_input

    def split(self, data: InputOutputData) -> Tuple[InputOutputData, InputOutputData]:
        if not isinstance(data, InputOutputData):
            raise ValueError(f"{self} is only applicable to instances of {InputOutputData.__name__}, got {data}")
        df = data.inputs if self.applyToInput else data.outputs
        indices_a, indices_b = self.dataFrameSplitter.compute_split_indices(df, self.fractionalSizeOfFirstSet)
        a = data.filter_indices(list(indices_a))
        b = data.filter_indices(list(indices_b))
        return a, b


class DataSplitterFromSkLearnSplitter(DataSplitter):
    def __init__(self, sklearn_splitter):
        """
        :param sklearn_splitter: an instance of one of the splitter classes from sklearn.model_selection,
            see https://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection
        """
        self.sklearn_splitter = sklearn_splitter

    def split(self, data: TInputOutputData) -> Tuple[TInputOutputData, TInputOutputData]:
        splitter_result = self.sklearn_splitter.split(data.inputs, data.outputs)
        split = next(iter(splitter_result))
        first_indices, second_indices = split
        return data.filter_indices(first_indices), data.filter_indices(second_indices)


class DataSplitterStratifiedShuffleSplit(DataSplitterFromSkLearnSplitter):
    def __init__(self, fractional_size_of_first_set: float, random_seed=42):
        super().__init__(StratifiedShuffleSplit(n_splits=1, train_size=fractional_size_of_first_set, random_state=random_seed))

    @staticmethod
    def is_applicable(io_data: InputOutputData):
        class_counts = io_data.outputs.value_counts()
        return all(class_counts >= 2)


class DataFrameSplitter(ABC):
    @abstractmethod
    def compute_split_indices(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[Sequence[int], Sequence[int]]:
        pass

    @staticmethod
    def split_with_indices(df: pd.DataFrame, indices_pair: Tuple[Sequence[int], Sequence[int]]) -> Tuple[pd.DataFrame, pd.DataFrame]:
        indices_a, indices_b = indices_pair
        a = df.iloc[indices_a]
        b = df.iloc[indices_b]
        return a, b

    def split(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[pd.DataFrame, pd.DataFrame]:
        return self.split_with_indices(df, self.compute_split_indices(df, fractional_size_of_first_set))


class DataFrameSplitterFractional(DataFrameSplitter):
    def __init__(self, shuffle=False, random_seed=42):
        self.randomSeed = random_seed
        self.shuffle = shuffle

    def compute_split_indices(self, df: pd.DataFrame, fractional_size_of_first_set: float) -> Tuple[Sequence[int], Sequence[int]]:
        n = df.shape[0]
        size_a = int(n * fractional_size_of_first_set)
        if self.shuffle:
            rand = np.random.RandomState(self.randomSeed)
            indices = rand.permutation(n)
        else:
            indices = list(range(n))
        indices_a = indices[:size_a]
        indices_b = indices[size_a:]
        return indices_a, indices_b



import functools
import logging
from abc import ABC, abstractmethod
from typing import Tuple, Dict, Any, Generator, Generic, TypeVar, Sequence, Optional, List, Union, Callable

import pandas as pd

from .eval_stats import GUESS
from .eval_stats.eval_stats_base import EvalStats, EvalStatsCollection
from .eval_stats.eval_stats_classification import ClassificationEvalStats, ClassificationMetric
from .eval_stats.eval_stats_regression import RegressionEvalStats, RegressionEvalStatsCollection, RegressionMetric
from ..data import DataSplitter, DataSplitterFractional, InputOutputData
from ..data_transformation import DataFrameTransformer
from ..tracking import TrackingMixin, TrackedExperiment
from ..util.deprecation import deprecated
from ..util.string import ToStringMixin
from ..util.typing import PandasNamedTuple
from ..vector_model import VectorClassificationModel, VectorModel, VectorModelBase, VectorModelFittableBase, VectorRegressionModel

log = logging.getLogger(__name__)

TModel = TypeVar("TModel", bound=VectorModel)
TEvalStats = TypeVar("TEvalStats", bound=EvalStats)
TEvalStatsCollection = TypeVar("TEvalStatsCollection", bound=EvalStatsCollection)


class MetricsDictProvider(TrackingMixin, ABC):
    @abstractmethod
    def _compute_metrics(self, model, **kwargs) -> Dict[str, float]:
        """
        Computes metrics for the given model, typically by fitting the model and applying it to test data

        :param model: the model
        :param kwargs: parameters to pass on to the underlying evaluation method
        :return: a dictionary with metrics values
        """
        pass

    def compute_metrics(self, model, **kwargs) -> Optional[Dict[str, float]]:
        """
        Computes metrics for the given model, typically by fitting the model and applying it to test data.
        If a tracked experiment was previously set, the metrics are tracked with the string representation
        of the model added under an additional key 'str(model)'.

        :param model: the model for which to compute metrics
        :param kwargs: parameters to pass on to the underlying evaluation method
        :return: a dictionary with metrics values
        """
        values_dict = self._compute_metrics(model, **kwargs)
        if self.tracked_experiment is not None:
            self.tracked_experiment.track_values(values_dict, add_values_dict={"str(model)": str(model)})  # TODO strings unsupported (mlflow)
        return values_dict


class MetricsDictProviderFromFunction(MetricsDictProvider):
    def __init__(self, compute_metrics_fn: Callable[[VectorModel], Dict[str, float]]):
        self._compute_metrics_fn = compute_metrics_fn

    def _compute_metrics(self, model, **kwargs) -> Dict[str, float]:
        return self._compute_metrics_fn(model)


class VectorModelEvaluationData(ABC, Generic[TEvalStats]):
    def __init__(self, stats_dict: Dict[str, TEvalStats], io_data: InputOutputData, model: VectorModelBase):
        """
        :param stats_dict: a dictionary mapping from output variable name to the evaluation statistics object
        :param io_data: the input/output data that was used to produce the results
        :param model: the model that was used to produce predictions
        """
        self.io_data = io_data
        self.eval_stats_by_var_name = stats_dict
        self.predicted_var_names = list(self.eval_stats_by_var_name.keys())
        self.model = model

    @property
    def model_name(self):
        return self.model.get_name()

    @property
    def input_data(self):  # for backward compatibility
        return self.io_data.inputs

    def get_eval_stats(self, predicted_var_name=None) -> TEvalStats:
        if predicted_var_name is None:
            if len(self.eval_stats_by_var_name) != 1:
                raise Exception(f"Must provide name of predicted variable name, as multiple variables were predicted:"
                    f" {list(self.eval_stats_by_var_name.keys())}")
            else:
                predicted_var_name = next(iter(self.eval_stats_by_var_name.keys()))
        eval_stats = self.eval_stats_by_var_name.get(predicted_var_name)
        if eval_stats is None:
            raise ValueError(f"No evaluation data present for '{predicted_var_name}'; known output variables: "
                f"{list(self.eval_stats_by_var_name.keys())}")
        return eval_stats

    def get_data_frame(self):
        """
        Returns an DataFrame with all evaluation metrics (one row per output variable)

        :return: a DataFrame containing evaluation metrics
        """
        stats_dicts = []
        var_names = []
        for predictedVarName, evalStats in self.eval_stats_by_var_name.items():
            stats_dicts.append(evalStats.metrics_dict())
            var_names.append(predictedVarName)
        df = pd.DataFrame(stats_dicts, index=var_names)
        df.index.name = "predictedVar"
        return df

    def iter_input_output_ground_truth_tuples(self, predicted_var_name=None) -> Generator[Tuple[PandasNamedTuple, Any, Any], None, None]:
        eval_stats = self.get_eval_stats(predicted_var_name)
        for i, named_tuple in enumerate(self.input_data.itertuples()):
            yield named_tuple, eval_stats.y_predicted[i], eval_stats.y_true[i]


class VectorRegressionModelEvaluationData(VectorModelEvaluationData[RegressionEvalStats]):
    def get_eval_stats_collection(self):
        return RegressionEvalStatsCollection(list(self.eval_stats_by_var_name.values()))


TEvalData = TypeVar("TEvalData", bound=VectorModelEvaluationData)


class EvaluatorParams(ToStringMixin, ABC):
    def __init__(self, data_splitter: DataSplitter = None, fractional_split_test_fraction: float = None, fractional_split_random_seed=42,
            fractional_split_shuffle=True):
        """
        :param data_splitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractional_split_test_fraction: [if test data must be obtained via split, dataSplitter is None] the fraction of the data to
            use for testing/evaluation;
        :param fractional_split_random_seed: [if test data must be obtained via split, dataSplitter is none] the random seed to use for the
            fractional split of the data
        :param fractional_split_shuffle: [if test data must be obtained via split, dataSplitter is None] whether to randomly (based on
            randomSeed) shuffle the dataset before splitting it
        """
        self._dataSplitter = data_splitter
        self._fractionalSplitTestFraction = fractional_split_test_fraction
        self._fractionalSplitRandomSeed = fractional_split_random_seed
        self._fractionalSplitShuffle = fractional_split_shuffle

    def _tostring_exclude_private(self) -> bool:
        return True

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        d = {}
        if self._dataSplitter is not None:
            d["dataSplitter"] = self._dataSplitter
        else:
            d["fractionalSplitTestFraction"] = self._fractionalSplitTestFraction
            d["fractionalSplitRandomSeed"] = self._fractionalSplitRandomSeed
            d["fractionalSplitShuffle"] = self._fractionalSplitShuffle
        return d

    def get_data_splitter(self) -> DataSplitter:
        if self._dataSplitter is None:
            if self._fractionalSplitTestFraction is None:
                raise ValueError("Cannot create default data splitter, as no split fraction was provided")
            self._dataSplitter = DataSplitterFractional(1 - self._fractionalSplitTestFraction, shuffle=self._fractionalSplitShuffle,
                random_seed=self._fractionalSplitRandomSeed)
        return self._dataSplitter

    def set_data_splitter(self, splitter: DataSplitter):
        self._dataSplitter = splitter


class VectorModelEvaluator(MetricsDictProvider, Generic[TEvalData], ABC):
    def __init__(self, data: Optional[InputOutputData], test_data: InputOutputData = None, params: EvaluatorParams = None):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param test_data: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        """
        if test_data is None:
            if params is None:
                raise ValueError("Parameters required for data split must be provided")
            data_splitter = params.get_data_splitter()
            self.training_data, self.test_data = data_splitter.split(data)
            log.debug(f"{data_splitter} created split with {len(self.training_data)} "
                f"({100 * len(self.training_data) / len(data):.2f}%) and "
                f"{len(self.test_data)} ({100 * len(self.test_data) / len(data):.2f}%) training and test data points respectively")
        else:
            self.training_data = data
            self.test_data = test_data

    def set_tracked_experiment(self, tracked_experiment: TrackedExperiment):
        """
        Sets a tracked experiment which will result in metrics being saved whenever computeMetrics is called
        or evalModel is called with track=True.

        :param tracked_experiment: the experiment in which to track evaluation metrics.
        """
        super().set_tracked_experiment(tracked_experiment)

    def eval_model(self, model: Union[VectorModelBase, VectorModelFittableBase], on_training_data=False, track=True,
            fit=False) -> TEvalData:
        """
        Evaluates the given model

        :param model: the model to evaluate
        :param on_training_data: if True, evaluate on this evaluator's training data rather than the held-out test data
        :param track: whether to track the evaluation metrics for the case where a tracked experiment was set on this object
        :param fit: whether to fit the model before evaluating it (via this object's `fit_model` method); if enabled, the model
            must support fitting
        :return: the evaluation result
        """
        data = self.training_data if on_training_data else self.test_data
        with self.begin_optional_tracking_context_for_model(model, track=track) as trackingContext:
            if fit:
                self.fit_model(model)
            result: VectorModelEvaluationData = self._eval_model(model, data)
            is_multiple_pred_vars = len(result.predicted_var_names) > 1
            for pred_var_name in result.predicted_var_names:
                metrics = result.get_eval_stats(pred_var_name).metrics_dict()
                trackingContext.track_metrics(metrics, pred_var_name if is_multiple_pred_vars else None)
        return result

    @abstractmethod
    def _eval_model(self, model: VectorModelBase, data: InputOutputData) -> TEvalData:
        pass

    def _compute_metrics(self, model: VectorModel, on_training_data=False) -> Dict[str, float]:
        return self._compute_metrics_for_var_name(model, None, on_training_data=on_training_data)

    def _compute_metrics_for_var_name(self, model, predicted_var_name: Optional[str], on_training_data=False):
        self.fit_model(model)
        track = False  # avoid duplicate tracking (as this function is only called by computeMetrics, which already tracks)
        eval_data: VectorModelEvaluationData = self.eval_model(model, on_training_data=on_training_data, track=track)
        return eval_data.get_eval_stats(predicted_var_name=predicted_var_name).metrics_dict()

    def create_metrics_dict_provider(self, predicted_var_name: Optional[str]) -> MetricsDictProvider:
        """
        Creates a metrics dictionary provider, e.g. for use in hyperparameter optimisation

        :param predicted_var_name: the name of the predicted variable for which to obtain evaluation metrics; may be None only
            if the model outputs but a single predicted variable
        :return: a metrics dictionary provider instance for the given variable
        """
        return MetricsDictProviderFromFunction(functools.partial(self._compute_metrics_for_var_name, predictedVarName=predicted_var_name))

    def fit_model(self, model: VectorModelFittableBase):
        """Fits the given model's parameters using this evaluator's training data"""
        if self.training_data is None:
            raise Exception(f"Cannot fit model with evaluator {self.__class__.__name__}: no training data provided")
        model.fit(self.training_data.inputs, self.training_data.outputs)


class RegressionEvaluatorParams(EvaluatorParams):
    def __init__(self,
            data_splitter: DataSplitter = None,
            fractional_split_test_fraction: float = None,
            fractional_split_random_seed=42,
            fractional_split_shuffle=True,
            metrics: Sequence[RegressionMetric] = None,
            additional_metrics: Sequence[RegressionMetric] = None,
            output_data_frame_transformer: DataFrameTransformer = None):
        """
        :param data_splitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractional_split_test_fraction: [if dataSplitter is None, test data must be obtained via split] the fraction of the data to
            use for testing/evaluation;
        :param fractional_split_random_seed: [if dataSplitter is none, test data must be obtained via split] the random seed to use for the
            fractional split of the data
        :param fractional_split_shuffle: [if dataSplitter is None, test data must be obtained via split] whether to randomly (based on
            randomSeed) shuffle the dataset before splitting it

        :param additional_metrics: additional regression metrics to apply
        :param output_data_frame_transformer: a data frame transformer to apply to all output data frames (both model outputs and ground
            truth), such that evaluation metrics are computed on the transformed data frame
        """
        super().__init__(data_splitter,
            fractional_split_test_fraction=fractional_split_test_fraction,
            fractional_split_random_seed=fractional_split_random_seed,
            fractional_split_shuffle=fractional_split_shuffle)
        self.metrics = metrics
        self.additional_metrics = additional_metrics
        self.output_data_frame_transformer = output_data_frame_transformer

    @classmethod
    def from_dict_or_instance(cls,
            params: Optional[Union[Dict[str, Any], "RegressionEvaluatorParams"]]) -> "RegressionEvaluatorParams":
        if params is None:
            return RegressionEvaluatorParams()
        elif type(params) == dict:
            raise Exception("Old-style dictionary parametrisation is no longer supported")
        elif isinstance(params, cls):
            return params
        else:
            raise ValueError(f"Must provide dictionary or {cls} instance, got {params}, type {type(params)}")


class VectorRegressionModelEvaluatorParams(RegressionEvaluatorParams):
    @deprecated("Use RegressionEvaluatorParams instead")
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VectorRegressionModelEvaluator(VectorModelEvaluator[VectorRegressionModelEvaluationData]):
    def __init__(self, data: Optional[InputOutputData], test_data: InputOutputData = None,
            params: RegressionEvaluatorParams = None):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param test_data: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        """
        # TODO params should not be optional
        super().__init__(data=data, test_data=test_data, params=params)
        self.params = params

    def _eval_model(self, model: VectorRegressionModel, data: InputOutputData) -> VectorRegressionModelEvaluationData:
        if not model.is_regression_model():
            raise ValueError(f"Expected a regression model, got {model}")
        eval_stats_by_var_name = {}
        predictions, ground_truth = self._compute_outputs(model, data)
        for predictedVarName in predictions.columns:
            eval_stats = RegressionEvalStats(y_predicted=predictions[predictedVarName], y_true=ground_truth[predictedVarName],
                metrics=self.params.metrics,
                additional_metrics=self.params.additional_metrics,
                model=model,
                io_data=data)
            eval_stats_by_var_name[predictedVarName] = eval_stats
        return VectorRegressionModelEvaluationData(eval_stats_by_var_name, data, model)

    def compute_test_data_outputs(self, model: VectorModelBase) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a pair (predictions, groundTruth)
        """
        return self._compute_outputs(model, self.test_data)

    def _compute_outputs(self, model: VectorModelBase, io_data: InputOutputData):
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param io_data: the data set
        :return: a pair (predictions, ground_truth)
        """
        predictions = model.predict(io_data.inputs)
        ground_truth = io_data.outputs
        if self.params.output_data_frame_transformer:
            predictions = self.params.output_data_frame_transformer.apply(predictions)
            ground_truth = self.params.output_data_frame_transformer.apply(ground_truth)
        return predictions, ground_truth


class VectorClassificationModelEvaluationData(VectorModelEvaluationData[ClassificationEvalStats]):
    def get_misclassified_inputs_data_frame(self) -> pd.DataFrame:
        return self.input_data.iloc[self.get_eval_stats().get_misclassified_indices()]

    def get_misclassified_triples_pred_true_input(self) -> List[Tuple[Any, Any, pd.Series]]:
        """
        :return: a list containing a triple (predicted class, true class, input series) for each misclassified data point
        """
        eval_stats = self.get_eval_stats()
        indices = eval_stats.get_misclassified_indices()
        return [(eval_stats.y_predicted[i], eval_stats.y_true[i], self.input_data.iloc[i]) for i in indices]


class ClassificationEvaluatorParams(EvaluatorParams):
    def __init__(self, data_splitter: DataSplitter = None, fractional_split_test_fraction: float = None, fractional_split_random_seed=42,
            fractional_split_shuffle=True, additional_metrics: Sequence[ClassificationMetric] = None,
            compute_probabilities: bool = False, binary_positive_label=GUESS):
        """
        :param data_splitter: [if test data must be obtained via split] a splitter to use in order to obtain; if None, must specify
            fractionalSplitTestFraction for fractional split (default)
        :param fractional_split_test_fraction: [if dataSplitter is None, test data must be obtained via split] the fraction of the data to
            use for testing/evaluation
        :param fractional_split_random_seed: [if dataSplitter is none, test data must be obtained via split] the random seed to use for the
            fractional split of the data
        :param fractional_split_shuffle: [if dataSplitter is None, test data must be obtained via split] whether to randomly (based on
            randomSeed) shuffle the dataset before splitting it
        :param additional_metrics: additional metrics to apply
        :param compute_probabilities: whether to compute class probabilities. Enabling this will enable many downstream computations
            and visualisations (e.g. precision-recall plots) but requires the model to support probability computation in general.
        :param binary_positive_label: the positive class label for binary classification; if GUESS, try to detect from labels;
            if None, no detection (assume non-binary classification)
        """
        super().__init__(data_splitter,
            fractional_split_test_fraction=fractional_split_test_fraction,
            fractional_split_random_seed=fractional_split_random_seed,
            fractional_split_shuffle=fractional_split_shuffle)
        self.additionalMetrics = additional_metrics
        self.computeProbabilities = compute_probabilities
        self.binaryPositiveLabel = binary_positive_label

    @classmethod
    def from_dict_or_instance(cls,
            params: Optional[Union[Dict[str, Any], "ClassificationEvaluatorParams"]]) \
            -> "ClassificationEvaluatorParams":
        if params is None:
            return ClassificationEvaluatorParams()
        elif type(params) == dict:
            raise ValueError("Old-style dictionary parametrisation is no longer supported")
        elif isinstance(params, ClassificationEvaluatorParams):
            return params
        else:
            raise ValueError(f"Must provide dictionary or instance, got {params}")


class VectorClassificationModelEvaluatorParams(ClassificationEvaluatorParams):
    @deprecated("Use ClassificationEvaluatorParams instead")
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class VectorClassificationModelEvaluator(VectorModelEvaluator[VectorClassificationModelEvaluationData]):
    def __init__(self,
            data: Optional[InputOutputData],
            test_data: InputOutputData = None,
            params: VectorClassificationModelEvaluatorParams = None):
        """
        Constructs an evaluator with test and training data.

        :param data: the full data set, or, if testData is given, the training data
        :param test_data: the data to use for testing/evaluation; if None, must specify appropriate parameters to define splitting
        :param params: the parameters
        """
        # TODO params should not be optional
        super().__init__(data=data, test_data=test_data, params=params)
        self.params = params

    def _eval_model(self, model: VectorClassificationModel, data: InputOutputData) -> VectorClassificationModelEvaluationData:
        if model.is_regression_model():
            raise ValueError(f"Expected a classification model, got {model}")
        predictions, predictions_proba, ground_truth = self._compute_outputs(model, data)
        eval_stats = ClassificationEvalStats(
            y_predicted_class_probabilities=predictions_proba,
            y_predicted=predictions,
            y_true=ground_truth,
            labels=model.get_class_labels(),
            additional_metrics=self.params.additionalMetrics,
            binary_positive_label=self.params.binaryPositiveLabel)
        predicted_var_name = model.get_predicted_variable_names()[0]
        return VectorClassificationModelEvaluationData({predicted_var_name: eval_stats}, data, model)

    def compute_test_data_outputs(self, model) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the test data

        :param model: the model to apply
        :return: a triple (predictions, predicted class probability vectors, groundTruth) of DataFrames
        """
        return self._compute_outputs(model, self.test_data)

    def _compute_outputs(self, model, io_data: InputOutputData) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Applies the given model to the given data

        :param model: the model to apply
        :param io_data: the data set
        :return: a triple (predictions, predicted class probability vectors, ground_truth) of DataFrames
        """
        if self.params.computeProbabilities:
            class_probabilities = model.predict_class_probabilities(io_data.inputs)
            predictions = model.convert_class_probabilities_to_predictions(class_probabilities)
        else:
            class_probabilities = None
            predictions = model.predict(io_data.inputs)
        ground_truth = io_data.outputs
        return predictions, class_probabilities, ground_truth


class RuleBasedVectorClassificationModelEvaluator(VectorClassificationModelEvaluator):
    def __init__(self, data: InputOutputData):
        super().__init__(data, test_data=data)

    def eval_model(self, model: VectorModelBase, on_training_data=False, track=True) -> VectorClassificationModelEvaluationData:
        """
        Evaluate the rule based model. The training data and test data coincide, thus fitting the model
        will fit the model's preprocessors on the full data set and evaluating it will evaluate the model on the
        same data set.

        :param model: the model to evaluate
        :param on_training_data: has to be False here. Setting to True is not supported and will lead to an
            exception
        :param track: whether to track the evaluation metrics for the case where a tracked experiment was set on this object
        :return: the evaluation result
        """
        if on_training_data:
            raise Exception("Evaluating rule based models on training data is not supported. In this evaluator"
                            "training and test data coincide.")
        return super().eval_model(model)


class RuleBasedVectorRegressionModelEvaluator(VectorRegressionModelEvaluator):
    def __init__(self, data: InputOutputData):
        super().__init__(data, test_data=data)

    def eval_model(self, model: VectorModelBase, on_training_data=False, track=True) -> VectorRegressionModelEvaluationData:
        """
        Evaluate the rule based model. The training data and test data coincide, thus fitting the model
        will fit the model's preprocessors on the full data set and evaluating it will evaluate the model on the
        same data set.

        :param model: the model to evaluate
        :param on_training_data: has to be False here. Setting to True is not supported and will lead to an
            exception
        :param track: whether to track the evaluation metrics for the case where a tracked experiment was set on this object
        :return: the evaluation result
        """
        if on_training_data:
            raise Exception("Evaluating rule based models on training data is not supported. In this evaluator"
                            "training and test data coincide.")
        return super().eval_model(model)



from typing import Sequence, Union, Optional
import logging
import pandas as pd
import re
import catboost

from .util.string import or_regex_group
from .sklearn.sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnVectorClassificationModel

log = logging.getLogger(__name__)


# noinspection DuplicatedCode
class CatBoostVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    log = log.getChild(__qualname__)

    def __init__(self, categorical_feature_names: Optional[Union[Sequence[str], str]] = None, random_state=42, num_leaves=31, **model_args):
        """
        :param categorical_feature_names: sequence of feature names in the input data that are categorical.
            Columns that have dtype 'category' (as will be the case for categorical columns created via FeatureGenerators)
            need not be specified (should be inferred automatically).
            In general, passing categorical features is preferable to using one-hot encoding, for example.
        :param random_state: the random seed to use
        :param num_leaves: the maximum number of leaves in one tree (original catboost default is 31)
        :param model_args: see https://catboost.ai/docs/concepts/python-reference_parameters-list.html#python-reference_parameters-list
        """
        super().__init__(catboost.CatBoostRegressor, random_seed=random_state, num_leaves=num_leaves, **model_args)

        if type(categorical_feature_names) == str:
            categorical_feature_name_regex = categorical_feature_names
        else:
            if categorical_feature_names is not None and len(categorical_feature_names) > 0:
                categorical_feature_name_regex = or_regex_group(categorical_feature_names)
            else:
                categorical_feature_name_regex = None
        self._categorical_feature_name_regex: str = categorical_feature_name_regex

    def _update_model_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        if self._categorical_feature_name_regex is not None:
            cols = list(inputs.columns)
            categorical_feature_names = [col for col in cols if re.match(self._categorical_feature_name_regex, col)]
            col_indices = [cols.index(f) for f in categorical_feature_names]
            args = {"cat_features": col_indices}
            self.log.info(f"Updating model parameters with {args}")
            self.modelArgs.update(args)


# noinspection DuplicatedCode
class CatBoostVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    log = log.getChild(__qualname__)

    def __init__(self, categorical_feature_names: Sequence[str] = None, random_state=42, num_leaves=31, **model_args):
        """
        :param categorical_feature_names: sequence of feature names in the input data that are categorical
            Columns that have dtype 'category' (as will be the case for categorical columns created via FeatureGenerators)
            need not be specified (should be inferred automatically, but we have never actually tested this behaviour
            successfully for a classification model).
            In general, passing categorical features may be preferable to using one-hot encoding, for example.
        :param random_state: the random seed to use
        :param num_leaves: the maximum number of leaves in one tree (original catboost default is 31)
        :param model_args: see https://catboost.ai/docs/concepts/python-reference_parameters-list.html#python-reference_parameters-list
        """
        super().__init__(catboost.CatBoostClassifier, random_seed=random_state, num_leaves=num_leaves, **model_args)

        if type(categorical_feature_names) == str:
            categorical_feature_name_regex = categorical_feature_names
        else:
            if categorical_feature_names is not None and len(categorical_feature_names) > 0:
                categorical_feature_name_regex = or_regex_group(categorical_feature_names)
            else:
                categorical_feature_name_regex = None
        self._categorical_feature_name_regex: str = categorical_feature_name_regex

    def _update_model_args(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
        if self._categorical_feature_name_regex is not None:
            cols = list(inputs.columns)
            categorical_feature_names = [col for col in cols if re.match(self._categorical_feature_name_regex, col)]
            col_indices = [cols.index(f) for f in categorical_feature_names]
            args = {"cat_features": col_indices}
            self.log.info(f"Updating model parameters with {args}")
            self.modelArgs.update(args)



import collections
import logging
import re
from typing import Optional, Union

import pandas as pd
import torch

from .lstnet_modules import LSTNetwork
from ...torch_base import TorchVectorClassificationModel, VectorTorchModel, ClassificationOutputMode
from ...torch_data import TorchDataSetProviderFromDataUtil, TensorScalerIdentity, TensorScaler, DataUtil
from ...torch_enums import ActivationFunction
from ...torch_opt import NNOptimiserParams
from ....util.string import object_repr

log: logging.Logger = logging.getLogger(__name__)


class LSTNetworkVectorClassificationModel(TorchVectorClassificationModel):
    """
    Classification model for time series data using the LSTNetwork architecture.

    Since the model takes a time series as input, it requires that input data frames to use special naming of columns
    such that the data can be interpreted correctly:
    Each column name must start with an N-digit prefix indicating the time slice the data pertains to (for any fixed N);
    the following suffix shall indicate the name of the actual feature.
    For each N-digit prefix, we must have the same set of suffixes in the list of columns, i.e. we must have the same
    features for each time slice in the input time series.
    """
    def __init__(self, num_input_time_slices, input_dim_per_time_slice, num_classes: Optional[int] = None,
            num_convolutions: int = 100, num_cnn_time_slices: int = 6, hid_rnn: int = 100, skip: int = 0, hid_skip: int = 5,
            hw_window: int = 0, hw_combine: str = "plus", dropout=0.2, output_activation=ActivationFunction.LOG_SOFTMAX, cuda=True,
            nn_optimiser_params: Union[dict, NNOptimiserParams] = None):
        """
        :param num_input_time_slices: the number of input time slices
        :param input_dim_per_time_slice: the dimension of the input data per time slice
        :param num_classes: the number of classes considered by this classification problem; if None, determine from data
        :param num_cnn_time_slices: the number of time slices considered by each convolution (i.e. it is one of the dimensions of the matrix
            used for convolutions, the other dimension being inputDimPerTimeSlice), a.k.a. "Ck"
        :param num_convolutions: the number of separate convolutions to apply, i.e. the number of independent convolution matrices,
            a.k.a "hidC";
            if it is 0, then the entire complex processing path is not applied.
        :param hid_rnn: the number of hidden output dimensions for the RNN stage
        :param skip: the number of time slices to skip for the skip-RNN. If it is 0, then the skip-RNN is not used.
        :param hid_skip: the number of output dimensions of each of the skip parallel RNNs
        :param hw_window: the number of time slices from the end of the input time series to consider as input for the highway component.
            If it is 0, the highway component is not used.
        :param hw_combine: {"plus", "product", "bilinear"} the function with which the highway component's output is combined with the
            complex path's output
        :param dropout: the dropout probability to use during training (dropouts are applied after every major step in the evaluation path)
        :param output_activation: the output activation function
        :param nn_optimiser_params: parameters for NNOptimiser to use for training
        """
        self.cuda = cuda
        self.numClasses = num_classes
        lstnet_args = dict(numInputTimeSlices=num_input_time_slices, inputDimPerTimeSlice=input_dim_per_time_slice, numOutputTimeSlices=1,
            outputDimPerTimeSlice=num_classes, numConvolutions=num_convolutions, numCnnTimeSlices=num_cnn_time_slices, hidRNN=hid_rnn,
            hwWindow=hw_window, hwCombine=hw_combine, dropout=dropout, outputActivation=output_activation,
            skip=skip, hidSkip=hid_skip, isClassification=True)
        output_mode = ClassificationOutputMode.for_activation_fn(ActivationFunction.torch_function_from_any(output_activation))
        super().__init__(output_mode, self._LSTNetworkModel, model_args=[self.cuda], model_kwargs=lstnet_args,
            nn_optimiser_params=nn_optimiser_params)

    class _LSTNetworkModel(VectorTorchModel):
        def __init__(self, cuda, **lstnet_args):
            """
            :param cuda: flag indicating whether cuda is used
            :param inputDim: the total number of inputs per data point
            :param numClasses: the number of classes to predict
            :param lstnet_args: arguments with which to construct the underlying LSTNetwork instance
            """
            super().__init__(cuda)
            self.lstnetArgs = lstnet_args

        def create_torch_module_for_dims(self, input_dim, output_dim):
            expected_input_dim = self.lstnetArgs["numInputTimeSlices"] * self.lstnetArgs["inputDimPerTimeSlice"]
            if expected_input_dim != input_dim:
                raise ValueError(f"Unexpected input size {input_dim}, expected {self.inputDim}")
            if self.lstnetArgs["outputDimPerTimeSlice"] is None:
                self.lstnetArgs["outputDimPerTimeSlice"] = output_dim
            else:
                if self.lstnetArgs["outputDimPerTimeSlice"] != output_dim:
                    raise ValueError(f"Unexpected output size {output_dim}, expected {self.lstnetArgs['outputDimPerTimeSlice']}")
            return LSTNetwork(**self.lstnetArgs)

        def __str__(self):
            return object_repr(self, self.lstnetArgs)

    def _create_data_set_provider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProviderFromDataUtil:
        if self.numClasses is None:
            self.numClasses = len(self._labels)
        elif self.numClasses != len(self._labels):
            raise ValueError(f"Output dimension {self.numClasses} per time time slice was specified, while the training data contains "
                             f"{len(self._labels)} classes")
        return TorchDataSetProviderFromDataUtil(self.DataUtil(inputs, outputs, self.numClasses), self.cuda)

    def _predict_outputs_for_input_data_frame(self, inputs: pd.DataFrame) -> torch.Tensor:
        log.info(f"Predicting outputs for {len(inputs)} inputs")
        result = super()._predict_outputs_for_input_data_frame(inputs)
        return result.squeeze(2)

    def _compute_model_inputs(self, x: pd.DataFrame, y: pd.DataFrame = None, fit=False) -> pd.DataFrame:
        x = super()._compute_model_inputs(x, y=y, fit=fit)

        # sort input data frame columns by name
        x = x[sorted(x.columns)]

        # check input column name format and consistency
        col_name_regex = re.compile(r"(\d+).+")
        cols_by_time_slice = collections.defaultdict(list)
        num_digits = None
        for colName in x.columns:
            match = col_name_regex.fullmatch(colName)
            if not match:
                raise ValueError(f"Column name '{colName}' does not match the required format (N-digit prefix indicating the time slice "
                                 f"order followed by feature name; for any fixed N); columns={list(x.columns)}")
            time_slice = match.group(1)
            if num_digits is None:
                num_digits = len(time_slice)
            elif num_digits != len(time_slice):
                raise ValueError(f"Inconsistent number of digits in column names: Got {num_digits} leading digits for one feature and "
                                 f"{len(time_slice)} for another; columns={list(x.columns)}")
            cols_by_time_slice[time_slice].append(colName[num_digits:])
        reference_cols = None
        for time_slice, cols in cols_by_time_slice.items():
            if reference_cols is None:
                reference_cols = cols
            elif reference_cols != cols:
                raise ValueError(f"Inconsistent features across time slices: Got suffixes {cols} for one time slice and {reference_cols} "
                                 f"for another; columns={list(x.columns)}")

        return x

    class DataUtil(DataUtil):
        def __init__(self, x_data: pd.DataFrame, y_data: pd.DataFrame, num_classes):
            self.y_data = y_data
            self.x_data = x_data
            self.numClasses = num_classes
            self.scaler = TensorScalerIdentity()

        def input_dim(self):
            return len(self.x_data.columns)

        def model_output_dim(self) -> int:
            return self.numClasses

        def split_into_tensors(self, fractional_size_of_first_set):
            split_index = round(fractional_size_of_first_set * len(self.y_data))
            y1, x1 = self.get_input_output_pair(self.y_data[:split_index], self.x_data[:split_index])
            y2, x2 = self.get_input_output_pair(self.y_data[split_index:], self.x_data[split_index:])
            return (x1, y1), (x2, y2)

        def get_input_output_pair(self, output, input):
            y = torch.tensor(output.values).long()
            x = torch.tensor(input.values).float()
            return y, x

        def get_output_tensor_scaler(self) -> TensorScaler:
            return self.scaler

        def get_input_tensor_scaler(self) -> TensorScaler:
            return self.scaler



import logging
from copy import copy

import numpy as np
import pandas as pd

log = logging.getLogger(__name__)


class DataFrameColumnChangeTracker:
    """
    A simple class for keeping track of changes in columns between an initial data frame and some other data frame
    (usually the result of some transformations performed on the initial one).

    Example:

    >>> from sensai.util.pandas import DataFrameColumnChangeTracker
    >>> import pandas as pd

    >>> df = pd.DataFrame({"bar": [1, 2]})
    >>> columnChangeTracker = DataFrameColumnChangeTracker(df)
    >>> df["foo"] = [4, 5]
    >>> columnChangeTracker.track_change(df)
    >>> columnChangeTracker.get_removed_columns()
    set()
    >>> columnChangeTracker.get_added_columns()
    {'foo'}
    """
    def __init__(self, initial_df: pd.DataFrame):
        self.initialColumns = copy(initial_df.columns)
        self.final_columns = None

    def track_change(self, changed_df: pd.DataFrame):
        self.final_columns = copy(changed_df.columns)

    def get_removed_columns(self):
        self.assert_change_was_tracked()
        return set(self.initialColumns).difference(self.final_columns)

    def get_added_columns(self):
        """
        Returns the columns in the last entry of the history that were not present the first one
        """
        self.assert_change_was_tracked()
        return set(self.final_columns).difference(self.initialColumns)

    def column_change_string(self):
        """
        Returns a string representation of the change
        """
        self.assert_change_was_tracked()
        if list(self.initialColumns) == list(self.final_columns):
            return "none"
        removed_cols, added_cols = self.get_removed_columns(), self.get_added_columns()
        if removed_cols == added_cols == set():
            return f"reordered {list(self.final_columns)}"

        return f"added={list(added_cols)}, removed={list(removed_cols)}"

    def assert_change_was_tracked(self):
        if self.final_columns is None:
            raise Exception(f"No change was tracked yet. "
                            f"Did you forget to call trackChange on the resulting data frame?")


def extract_array(df: pd.DataFrame, dtype=None):
    """
    Extracts array from data frame. It is expected that each row corresponds to a data point and
    each column corresponds to a "channel". Moreover, all entries are expected to be arrays of the same shape
    (or scalars or sequences of the same length). We will refer to that shape as tensorShape.

    The output will be of shape `(N_rows, N_columns, *tensorShape)`. Thus, `N_rows` can be interpreted as dataset length
    (or batch size, if a single batch is passed) and N_columns can be interpreted as number of channels.
    Empty dimensions will be stripped, thus if the data frame has only one column, the array will have shape
    `(N_rows, *tensorShape)`.
    E.g. an image with three channels could equally be passed as data frame of the type


    +------------------+------------------+------------------+
    | R                | G                | B                |
    +==================+==================+==================+
    | channel          | channel          | channel          |
    +------------------+------------------+------------------+
    | channel          | channel          | channel          |
    +------------------+------------------+------------------+
    | ...              | ...              | ...              |
    +------------------+------------------+------------------+

    or as data frame of type

    +------------------+
    | image            |
    +==================+
    | RGB-array        |
    +------------------+
    | RGB-array        |
    +------------------+
    | ...              |
    +------------------+

    In both cases the returned array will have shape `(N_images, 3, width, height)`

    :param df: data frame where each entry is an array of shape tensorShape
    :param dtype: if not None, convert the array's data type to this type (string or numpy dtype)
    :return: array of shape `(N_rows, N_columns, *tensorShape)` with stripped empty dimensions
    """
    log.debug(f"Stacking tensors of shape {np.array(df.iloc[0, 0]).shape}")
    try:
        # This compact way of extracting the array causes dtypes to be modified,
        #    arr = np.stack(df.apply(np.stack, axis=1)).squeeze()
        # so we use this numpy-only alternative:
        arr = df.values
        if arr.shape[1] > 1:
            arr = np.stack([np.stack(arr[i]) for i in range(arr.shape[0])])
        else:
            arr = np.stack(arr[:, 0])
        # For the case where there is only one row, the old implementation above removed the first dimension,
        # so we do the same, even though it seems odd to do so (potential problem for batch size 1)
        # TODO: remove this behavior
        if arr.shape[0] == 1:
            arr = arr[0]
    except ValueError:
        raise ValueError(f"No array can be extracted from frame of length {len(df)} with columns {list(df.columns)}. "
                         f"Make sure that all entries have the same shape")
    if dtype is not None:
        arr = arr.astype(dtype, copy=False)
    return arr


def remove_duplicate_index_entries(df: pd.DataFrame):
    """
    Removes successive duplicate index entries by keeping only the first occurrence for every duplicate index element.

    :param df: the data frame, which is assumed to have a sorted index
    :return: the (modified) data frame with duplicate index entries removed
    """
    keep = [True]
    prev_item = df.index[0]
    for item in df.index[1:]:
        keep.append(item != prev_item)
        prev_item = item
    return df[keep]



"""
Utility functions and classes for geographic coordinates
"""

import math
from typing import Tuple, List, Generator

from ._globalmaptiles import GlobalMercator
from .geo_coords import GeoRect

EARTH_RADIUS = 6371000
EARTH_CIRCUMFERENCE = 2 * math.pi * EARTH_RADIUS
LATITUDE_PER_METRE = 360.0 / EARTH_CIRCUMFERENCE


class MapTile:
    def __init__(self, tx: int, ty: int, rect: GeoRect, zoom: int):
        self.tx = tx
        self.ty = ty
        self.rect = rect
        self.zoom = zoom


class MapTiles:
    def __init__(self, zoom=13):
        self.zoom = zoom
        self._mercator = GlobalMercator()
        self._tiles = {}

    def _get_tile(self, tx, ty):
        key = (tx, ty)
        tile = self._tiles.get(key)
        if tile is None:
            tile = MapTile(tx, ty, GeoRect(*self._mercator.TileLatLonBounds(tx, ty, self.zoom)), self.zoom)
            self._tiles[key] = tile
        return tile

    def iter_tile_coordinates_in_rect(self, rect: GeoRect) -> Generator[Tuple[int, int], None, None]:
        tx1, ty1 = self._mercator.LatLonToTile(rect.minLat, rect.minLon, self.zoom)
        tx2, ty2 = self._mercator.LatLonToTile(rect.maxLat, rect.maxLon, self.zoom)
        tx_min = min(tx1, tx2)
        tx_max = max(tx1, tx2)
        ty_min = min(ty1, ty2)
        ty_max = max(ty1, ty2)
        for tx in range(tx_min, tx_max+1):
            for ty in range(ty_min, ty_max+1):
                yield tx, ty

    def get_tiles_in_rect(self, rect: GeoRect) -> List[MapTile]:
        return [self._get_tile(tx, ty) for tx, ty in self.iter_tile_coordinates_in_rect(rect)]

    def get_tile(self, lat: float, lon: float) -> MapTile:
        return self._get_tile(*self.get_tile_coordinates(lat, lon))

    def get_tile_coordinates(self, lat: float, lon: float) -> Tuple[int, int]:
        return self._mercator.LatLonToTile(lat, lon, self.zoom)



import functools
from typing import Union

import pandas as pd

from . import VectorRegressionModel
from .vector_model import RuleBasedVectorRegressionModel


class MultiVectorRegressionModel(RuleBasedVectorRegressionModel):
    """
    Combines several (previously trained) regression models into a single regression model that produces the combined output of the
    individual models (concatenating their outputs)
    """
    def __init__(self, *models: VectorRegressionModel):
        self.models = models
        predicted_variable_names_list = [m.get_predicted_variable_names() for m in models]
        predicted_variable_names = functools.reduce(lambda x, y: x + y.get_predicted_variable_names(), models, [])
        if len(predicted_variable_names) != sum((len(v) for v in predicted_variable_names_list)):
            raise ValueError(f"Models do not produce disjoint outputs: {predicted_variable_names_list}")
        super().__init__(predicted_variable_names)

    def _predict(self, x: pd.DataFrame) -> Union[pd.DataFrame, list]:
        dfs = [m.predict(x) for m in self.models]
        combined_df = pd.concat(dfs, axis=1)
        return combined_df


import sys

from . import aggregation
from . import cache
from . import cache_mysql
from .dtype import dict_to_ordered_tuples
from .sequences import concat_sequences
from .helper import *
from .logging import LogTime


def _backward_compatibility():
    # module 'counter' was moved to 'aggregation'
    sys.modules[f"{__name__}.counter"] = aggregation


_backward_compatibility()



import logging
from typing import Union, Optional

import sklearn.ensemble
import sklearn.linear_model
import sklearn.neighbors
import sklearn.neural_network
import sklearn.svm
from matplotlib import pyplot as plt

from .sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnMultiDimVectorRegressionModel, \
    FeatureImportanceProviderSkLearnRegressionMultipleOneDim, FeatureImportanceProviderSkLearnRegressionMultiDim

log = logging.getLogger(__name__)


class SkLearnRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **model_args):
        super().__init__(sklearn.ensemble.RandomForestRegressor,
            n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **model_args)


class SkLearnLinearRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultiDim):
    def __init__(self, fit_intercept=True, **model_args):
        """
        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
            set to False if the data is already centred
        :param model_args: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
        """
        super().__init__(sklearn.linear_model.LinearRegression, fit_intercept=fit_intercept, **model_args)


class SkLearnLinearRidgeRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultiDim):
    """
    Linear least squares with L2 regularisation
    """
    def __init__(self, alpha=1.0, fit_intercept=True, solver="auto", max_iter=None, tol=1e-3, **model_args):
        """
        :param alpha: multiplies the L2 term, controlling regularisation strength
        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
            set to False if the data is already centred
        :param model_args: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html#sklearn.linear_model.Ridge
        """
        super().__init__(sklearn.linear_model.Ridge, alpha=alpha, fit_intercept=fit_intercept, max_iter=max_iter, tol=tol,
            solver=solver, **model_args)


class SkLearnLinearLassoRegressionVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultiDim):
    """
    Linear least squares with L1 regularisation, a.k.a. the lasso
    """
    def __init__(self, alpha=1.0, fit_intercept=True, max_iter=1000, tol=0.0001, **model_args):
        """
        :param alpha: multiplies the L1 term, controlling regularisation strength
        :param fit_intercept: whether to determine the intercept, i.e. the constant term which is not scaled with an input feature value;
            set to False if the data is already centred
        :param model_args: see https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html#sklearn.linear_model.Lasso
        """
        super().__init__(sklearn.linear_model.Lasso, alpha=alpha, fit_intercept=fit_intercept, max_iter=max_iter, tol=tol, **model_args)


class SkLearnMultiLayerPerceptronVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self,
            hidden_layer_sizes=(100,), activation: str = "relu",
            solver: str = "adam", batch_size: Union[int, str] = "auto", random_state: Optional[int] = 42,
            max_iter: int = 200, early_stopping: bool = False, n_iter_no_change: int = 10, **model_args):
        """
        :param hidden_layer_sizes: the sequence of hidden layer sizes
        :param activation: {"identity", "logistic", "tanh", "relu"} the activation function to use for hidden layers (the one used for the
            output layer is always 'identity')
        :param solver: {"adam", "lbfgs", "sgd"} the name of the solver to apply
        :param batch_size: the batch size or "auto" for min(200, data set size)
        :param random_state: the random seed for reproducability; use None if it shall not be specifically defined
        :param max_iter: the number of iterations (gradient steps for L-BFGS, epochs for other solvers)
        :param early_stopping: whether to use early stopping (stop training after n_iter_no_change epochs without improvement)
        :param n_iter_no_change: the number of iterations after which to stop early (if early_stopping is enabled)
        :param model_args: additional arguments to pass on to MLPRegressor,
            see https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html
        """
        super().__init__(sklearn.neural_network.MLPRegressor,
            random_state=random_state, hidden_layer_sizes=hidden_layer_sizes, activation=activation, solver=solver, batch_size=batch_size,
            max_iter=max_iter, early_stopping=early_stopping, n_iter_no_change=n_iter_no_change, **model_args)


class SkLearnSVRVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **model_args):
        super().__init__(sklearn.svm.SVR, **model_args)


class SkLearnLinearSVRVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **model_args):
        super().__init__(sklearn.svm.LinearSVR, **model_args)


class SkLearnGradientBoostingVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, random_state=42, **model_args):
        super().__init__(sklearn.ensemble.GradientBoostingRegressor, random_state=random_state, **model_args)


class SkLearnKNeighborsVectorRegressionModel(AbstractSkLearnMultiDimVectorRegressionModel):
    def __init__(self, **model_args):
        super().__init__(sklearn.neighbors.KNeighborsRegressor, **model_args)


class SkLearnExtraTreesVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, n_estimators=100, min_samples_leaf=10, random_state=42, **model_args):
        super().__init__(sklearn.ensemble.ExtraTreesRegressor,
            n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, random_state=random_state, **model_args)


class SkLearnDummyVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, strategy='mean', constant=None, quantile=None):
        super().__init__(sklearn.dummy.DummyRegressor,
            strategy=strategy, constant=constant, quantile=quantile)


class SkLearnDecisionTreeVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel):
    def __init__(self, random_state=42, **model_args):
        super().__init__(sklearn.tree.DecisionTreeRegressor, random_state=random_state, **model_args)

    def plot(self, predicted_var_name=None, figsize=None) -> plt.Figure:
        model = self.get_sklearn_model(predicted_var_name)
        fig = plt.figure(figsize=figsize)
        sklearn.tree.plot_tree(model, feature_names=self.get_model_input_variable_names())
        return fig

    def plot_graphviz_pdf(self, dot_path, predicted_var_name=None):
        """
        :param path: the path to a .dot file that will be created, alongside which a rendered PDF file (with added suffix ".pdf")
            will be placed
        :param predicted_var_name: the predicted variable name for which to plot the model (if multiple; None is admissible if
            there is only one predicted variable)
        """
        import graphviz
        dot = sklearn.tree.export_graphviz(self.get_sklearn_model(predicted_var_name), out_file=None,
            feature_names=self.get_model_input_variable_names(), filled=True)
        graphviz.Source(dot).render(dot_path)




import numpy as np
from pytorch_lightning import Trainer, LightningModule
from torch import tensor

from .. import tensor_model as tm
from ..data import InputOutputArrays, DataSplitterFractional


def _fit_model_with_trainer(model: LightningModule, trainer: Trainer, io_data,
        batch_size: int, splitter: DataSplitterFractional = None):
    if splitter is not None:
        train_io_data, validation_io_data = splitter.split(io_data)
        train_data_loader = train_io_data.toTorchDataLoader(batchSize=batch_size)
        val_data_loader = validation_io_data.toTorchDataLoader(batchSize=batch_size)
    else:
        train_data_loader = io_data.to_torch_data_loader(batch_size=batch_size)
        val_data_loader = None
    trainer.fit(model, train_data_loader, val_dataloaders=val_data_loader)


class PLWrappedModel:
    def __init__(self, model: LightningModule, trainer: Trainer, validation_fraction=0.1, shuffle=True, batch_size=32):
        if not 0 <= validation_fraction <= 1:
            raise ValueError(f"Invalid validationFraction: {validation_fraction}. Has to be in interval [0, 1]")
        self.trainer = trainer
        self.model = model
        self.validationFraction = validation_fraction
        self.shuffle = shuffle
        self.batchSize = batch_size

    def fit(self, x: np.ndarray, y: np.ndarray):
        io_data = InputOutputArrays(x, y)
        splitter = DataSplitterFractional(1 - self.validationFraction, shuffle=self.shuffle)
        _fit_model_with_trainer(self.model, self.trainer, io_data, self.batchSize, splitter=splitter)

    def predict(self, x: np.ndarray) -> np.ndarray:
        x = tensor(x)
        return self.model(x).detach().cpu().numpy()


# noinspection DuplicatedCode
class PLTensorToScalarClassificationModel(tm.TensorToScalarClassificationModel):
    def __init__(self, model: LightningModule, trainer: Trainer, validation_fraction=0.1, shuffle=True, batch_size=64,
                 check_input_shape=True, check_input_columns=True):
        super().__init__(check_input_shape=check_input_shape, check_input_columns=check_input_columns)
        self.wrapped_model = PLWrappedModel(model, trainer, validation_fraction=validation_fraction, shuffle=shuffle,
                                           batch_size=batch_size)

    def _predict_probabilities_array(self, x: np.ndarray) -> np.ndarray:
        return self.wrapped_model.predict(x)

    def _fit_to_array(self, x: np.ndarray, y: np.ndarray):
        self.wrapped_model.fit(x, y)


# noinspection DuplicatedCode
class PLTensorToScalarRegressionModel(tm.TensorToScalarRegressionModel):
    def __init__(self, model: LightningModule, trainer: Trainer, validation_fraction=0.1, shuffle=True, batch_size=32,
                 check_input_shape=True, check_input_columns=True):
        super().__init__(check_input_shape=check_input_shape, check_input_columns=check_input_columns)
        self.wrapped_model = PLWrappedModel(model, trainer, validation_fraction=validation_fraction, shuffle=shuffle,
                                           batch_size=batch_size)

    def _predict_array(self, x: np.ndarray) -> np.ndarray:
        return self.wrapped_model.predict(x)

    def _fit_to_array(self, x: np.ndarray, y: np.ndarray):
        self.wrapped_model.fit(x, y)


# noinspection DuplicatedCode
class PLTensorToTensorClassificationModel(tm.TensorToTensorClassificationModel):
    def __init__(self, model: LightningModule, trainer: Trainer, validation_fraction=0.1, shuffle=True, batch_size=32,
                 check_input_shape=True, check_input_columns=True):
        super().__init__(check_input_shape=check_input_shape, check_input_columns=check_input_columns)
        self.wrapped_model = PLWrappedModel(model, trainer, validation_fraction=validation_fraction, shuffle=shuffle,
                                           batch_size=batch_size)

    def _predict_probabilities_array(self, x: np.ndarray) -> np.ndarray:
        return self.wrapped_model.predict(x)

    def _fit_to_array(self, x: np.ndarray, y: np.ndarray):
        self.wrapped_model.fit(x, y)


# noinspection DuplicatedCode
class PLTensorToTensorRegressionModel(tm.TensorToTensorRegressionModel):
    def __init__(self, model: LightningModule, trainer: Trainer, validation_fraction=0.1, shuffle=True, batch_size=32,
                 check_input_shape=True, check_input_columns=True):
        super().__init__(check_input_shape=check_input_shape, check_input_columns=check_input_columns)
        self.wrapped_model = PLWrappedModel(model, trainer, validation_fraction=validation_fraction, shuffle=shuffle,
                                           batch_size=batch_size)

    def _predict_array(self, x: np.ndarray) -> np.ndarray:
        return self.wrapped_model.predict(x)

    def _fit_to_array(self, x: np.ndarray, y: np.ndarray):
        self.wrapped_model.fit(x, y)



import io
import logging
import os
from typing import Sequence, Optional, Tuple, List

import matplotlib.figure
from matplotlib import pyplot as plt
import pandas as pd

log = logging.getLogger(__name__)


class ResultWriter:
    log = log.getChild(__qualname__)

    def __init__(self, result_dir, filename_prefix=""):
        self.result_dir = result_dir
        os.makedirs(result_dir, exist_ok=True)
        self.filename_prefix = filename_prefix

    def child_with_added_prefix(self, prefix: str) -> "ResultWriter":
        """
        Creates a derived result writer with an added prefix, i.e. the given prefix is appended to this
        result writer's prefix

        :param prefix: the prefix to append
        :return: a new writer instance
        """
        return ResultWriter(self.result_dir, filename_prefix=self.filename_prefix + prefix)

    def child_for_subdirectory(self, dir_name: str):
        result_dir = os.path.join(self.result_dir, dir_name)
        os.makedirs(result_dir, exist_ok=True)
        return ResultWriter(result_dir, filename_prefix=self.filename_prefix)

    def path(self, filename_suffix: str, extension_to_add=None, valid_other_extensions: Optional[Sequence[str]] = None):
        """
        :param filename_suffix: the suffix to add (which may or may not already include a file extension)
        :param extension_to_add: if not None, the file extension to add (without the leading ".") unless
            the extension to add or one of the extenions in valid_extensions is already present
        :param valid_other_extensions: a sequence of valid other extensions (without the "."), only
            relevant if extensionToAdd is specified
        :return: the full path
        """
        if extension_to_add is not None:
            add_ext = True
            valid_extensions = set(valid_other_extensions) if valid_other_extensions is not None else set()
            valid_extensions.add(extension_to_add)
            if valid_extensions is not None:
                for ext in valid_extensions:
                    if filename_suffix.endswith("." + ext):
                        add_ext = False
                        break
            if add_ext:
                filename_suffix += "." + extension_to_add
        path = os.path.join(self.result_dir, f"{self.filename_prefix}{filename_suffix}")
        return path

    def write_text_file(self, filename_suffix, content):
        p = self.path(filename_suffix, extension_to_add="txt")
        self.log.info(f"Saving text file {p}")
        with open(p, "w") as f:
            f.write(content)
        return p

    def write_text_file_lines(self, filename_suffix, lines: List[str]):
        p = self.path(filename_suffix, extension_to_add="txt")
        self.log.info(f"Saving text file {p}")
        write_text_file_lines(lines, p)
        return p

    def write_data_frame_text_file(self, filename_suffix, df: pd.DataFrame):
        p = self.path(filename_suffix, extension_to_add="df.txt", valid_other_extensions="txt")
        self.log.info(f"Saving data frame text file {p}")
        with open(p, "w") as f:
            f.write(df.to_string())
        return p

    def write_data_frame_csv_file(self, filename_suffix, df: pd.DataFrame, index=True, header=True):
        p = self.path(filename_suffix, extension_to_add="csv")
        self.log.info(f"Saving data frame CSV file {p}")
        df.to_csv(p, index=index, header=header)
        return p

    def write_figure(self, filename_suffix, fig, close_figure=False):
        """
        :param filename_suffix: the filename suffix, which may or may not include a file extension, valid extensions being {"png", "jpg"}
        :param fig: the figure to save
        :param close_figure: whether to close the figure after having saved it
        :return: the path to the file that was written
        """
        p = self.path(filename_suffix, extension_to_add="png", valid_other_extensions=("jpg",))
        self.log.info(f"Saving figure {p}")
        fig.savefig(p)
        if close_figure:
            plt.close(fig)
        return p

    def write_figures(self, figures: Sequence[Tuple[str, matplotlib.figure.Figure]], close_figures=False):
        for name, fig in figures:
            self.write_figure(name, fig, close_figure=close_figures)

    def write_pickle(self, filename_suffix, obj):
        from .pickle import dump_pickle
        p = self.path(filename_suffix, extension_to_add="pickle")
        self.log.info(f"Saving pickle {p}")
        dump_pickle(obj, p)
        return p


def write_text_file_lines(lines: List[str], path):
    """
    :param lines: the lines to write (without a trailing newline, which will be added)
    :param path: the path of the text file to write to
    """
    with open(path, "w") as f:
        for line in lines:
            f.write(line)
            f.write("\n")


def read_text_file_lines(path, strip=True, skip_empty=True) -> List[str]:
    """
    :param path: the path of the text file to read from
    :param strip: whether to strip each line, removing whitespace/newline characters
    :param skip_empty: whether to skip any lines that are empty (after stripping)
    :return: the list of lines
    """
    lines = []
    with open(path, "r") as f:
        for line in f.readlines():
            if strip:
                line = line.strip()
            if not skip_empty or line != "":
                lines.append(line)
    return lines


def is_s3_path(path: str):
    return path.startswith("s3://")


class S3Object:
    def __init__(self, path):
        assert is_s3_path(path)
        self.path = path
        self.bucket, self.object = self.path[5:].split("/", 1)

    class OutputFile:
        def __init__(self, s3_object: "S3Object"):
            self.s3Object = s3_object
            self.buffer = io.BytesIO()

        def write(self, obj: bytes):
            self.buffer.write(obj)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.s3Object.put(self.buffer.getvalue())

    def get_file_content(self):
        return self._get_s3_object().get()['Body'].read()

    def open_file(self, mode):
        assert mode in ("wb", "rb")
        if mode == "rb":
            content = self.get_file_content()
            return io.BytesIO(content)

        elif mode == "wb":
            return self.OutputFile(self)

        else:
            raise ValueError(mode)

    def put(self, obj: bytes):
        self._get_s3_object().put(Body=obj)

    def _get_s3_object(self):
        import boto3
        session = boto3.session.Session(profile_name=os.getenv("AWS_PROFILE"))
        s3 = session.resource("s3")
        return s3.Bucket(self.bucket).Object(self.object)



from abc import ABC, abstractmethod
from concurrent.futures.process import ProcessPoolExecutor
from typing import Sequence, List
from inspect import currentframe, getframeinfo

import pandas as pd

from ..vector_model import VectorModel
from ..util.multiprocessing import VectorModelWithSeparateFeatureGeneration
from ..util.pickle import PickleFailureDebugger


class EnsembleVectorModel(VectorModel, ABC):
    def __init__(self, models: Sequence[VectorModel], num_processes=1):
        """
        :param models:
        :param num_processes:
        """
        self.num_processes = num_processes
        self.models = list(models)
        super().__init__(check_input_columns=False)

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame):
        if self.num_processes == 1 or len(self.models) == 1:
            for model in self.models:
                model.fit(x, y)
            return

        fitted_model_futures = []
        executor = ProcessPoolExecutor(max_workers=self.num_processes)
        fitters = [VectorModelWithSeparateFeatureGeneration(model) for model in self.models]
        for fitter in fitters:
            intermediate_step = fitter.fit_start(x, y)
            frame_info = getframeinfo(currentframe())
            PickleFailureDebugger.log_failure_if_enabled(intermediate_step,
                context_info=f"Submitting {fitter} in {frame_info.filename}:{frame_info.lineno}")
            fitted_model_futures.append(executor.submit(intermediate_step.execute))
        for i, fittedModelFuture in enumerate(fitted_model_futures):
            self.models[i] = fitters[i].fit_end(fittedModelFuture.result())

    def compute_all_predictions(self, x: pd.DataFrame):
        if self.num_processes == 1 or len(self.models) == 1:
            return [model.predict(x) for model in self.models]

        prediction_futures = []
        executor = ProcessPoolExecutor(max_workers=self.num_processes)
        predictors = [VectorModelWithSeparateFeatureGeneration(model) for model in self.models]
        for predictor in predictors:
            predict_finaliser = predictor.predict_start(x)
            frame_info = getframeinfo(currentframe())
            PickleFailureDebugger.log_failure_if_enabled(predict_finaliser,
                context_info=f"Submitting {predict_finaliser} in {frame_info.filename}:{frame_info.lineno}")
            prediction_futures.append(executor.submit(predict_finaliser.execute))
        return [predictionFuture.result() for predictionFuture in prediction_futures]

    def _predict(self, x):
        predictions_data_frames = self.compute_all_predictions(x)
        return self.aggregate_predictions(predictions_data_frames)

    @abstractmethod
    def aggregate_predictions(self, predictions_data_frames: List[pd.DataFrame]) -> pd.DataFrame:
        pass


class EnsembleRegressionVectorModel(EnsembleVectorModel, ABC):
    def is_regression_model(self):
        return True


class EnsembleClassificationVectorModel(EnsembleVectorModel, ABC):
    def is_regression_model(self):
        return False



from enum import Enum
from typing import Union

import numpy as np
import pandas as pd
import sklearn.preprocessing

from .util.dtype import to_float_array


class NormalisationMode(Enum):
    NONE = "none"
    MAX_ALL = "max_all"
    MAX_BY_COLUMN = "max_by_column"
    STANDARDISED = "standardised"


class VectorDataScaler:
    def __init__(self, data_frame: pd.DataFrame, normalisation_mode: NormalisationMode):
        self.normalisation_mode = normalisation_mode
        self.scale, self.translate = self._compute_scaling_params(data_frame.values, normalisation_mode)
        self.dimension_names = list(data_frame.columns)

    @classmethod
    def _compute_scaling_params(cls, raw_array: np.ndarray, normalisation_mode: NormalisationMode):
        """
        :param raw_array: numpy array containing raw data
        :param normalisation_mode: the normalization mode (0=none, 1=by maximum in entire data set, 2=by separate maximum in each column)
        """
        translate = None
        scale = None
        if normalisation_mode != NormalisationMode.NONE:
            if len(raw_array.shape) != 2:
                raise ValueError(f"Only 2D arrays are supported by {cls.__name__} with mode {normalisation_mode}")
            dim = raw_array.shape[1]
            if normalisation_mode == NormalisationMode.MAX_ALL:
                scale = np.ones(dim) * np.max(np.abs(raw_array))
            elif normalisation_mode == NormalisationMode.MAX_BY_COLUMN:
                scale = np.ones(dim)
                for i in range(dim):
                    scale[i] = np.max(np.abs(raw_array[:, i]))
            elif normalisation_mode == NormalisationMode.STANDARDISED:
                standardScaler = sklearn.preprocessing.StandardScaler()
                standardScaler.fit(raw_array)
                translate = standardScaler.mean_
                scale = standardScaler.scale_
            else:
                raise Exception("Unknown normalization mode")
        return scale, translate

    @staticmethod
    def _array(data: Union[pd.DataFrame, np.ndarray]):
        return to_float_array(data)

    def get_normalised_array(self, data: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
        result = self._array(data)
        if self.translate is not None:
            result = result - self.translate
        if self.scale is not None:
            result = result / self.scale
        return result

    def get_denormalised_array(self, data: Union[pd.DataFrame, np.ndarray]) -> np.ndarray:
        result = self._array(data)
        if self.scale is not None:
            result = result * self.scale
        if self.translate is not None:
            result = result + self.translate
        return result



import logging
from typing import Callable, Union, Iterable

import geopandas as gp
import numpy as np
import pandas as pd
from shapely.geometry import MultiPoint

from .coordinates import validate_coordinates, extract_coordinates_array, TCoordinates, GeoDataFrameWrapper
from ...clustering import SkLearnEuclideanClusterer
from ...clustering.clustering_base import EuclideanClusterer
from ...clustering.sklearn_clustering import SkLearnClustererProtocol
from ...util.cache import LoadSaveInterface
from ...util.profiling import timed

log = logging.getLogger(__name__)


class CoordinateEuclideanClusterer(EuclideanClusterer, GeoDataFrameWrapper):
    """
    Wrapper around a clustering model. This class adds additional, geospatial-specific features to the provided
    clusterer

    :param clusterer: an instance of ClusteringModel
    """
    def __init__(self, clusterer: EuclideanClusterer):
        self.clusterer = clusterer
        super().__init__(noise_label=clusterer.noiseLabel,
            max_cluster_size=clusterer.maxClusterSize, min_cluster_size=clusterer.minClusterSize)

    class Cluster(EuclideanClusterer.Cluster, GeoDataFrameWrapper, LoadSaveInterface):
        """
        Wrapper around a coordinates array

        :param coordinates:
        :param identifier:
        """

        def __init__(self, coordinates: np.ndarray, identifier: Union[str, int]):
            validate_coordinates(coordinates)
            super().__init__(coordinates, identifier)

        def to_geodf(self, crs='epsg:3857'):
            """
            Export the cluster as a GeoDataFrame of length 1 with the cluster as an instance of
            MultiPoint and the identifier as index.

            :param crs: projection. By default pseudo-mercator
            :return: GeoDataFrame
            """
            gdf = gp.GeoDataFrame({"geometry": [self.as_multipoint()]}, index=[self.identifier])
            gdf.index.name = "identifier"
            gdf.crs = crs
            return gdf

        def as_multipoint(self):
            """
            :return: The cluster's coordinates as a MultiPoint object
            """
            return MultiPoint(self.datapoints)

        @classmethod
        def load(cls, path):
            """
            Instantiate from a geopandas readable file containing a single row with an identifier and an instance
            of MultiPoint

            :param path:
            :return: instance of CoordinateCluster
            """
            log.info(f"Loading instance of {cls.__name__} from {path}")
            gdf = gp.read_file(path)
            if len(gdf) != 1:
                raise Exception(f"Expected {path} to contain a single row, instead got {len(gdf)}")
            identifier, multipoint = gdf.identifier.values[0], gdf.geometry.values[0]
            return cls(np.array([[p.x, p.y] for p in multipoint]), identifier)

        def save(self, path, crs="EPSG:3857"):
            """
            Saves the cluster's coordinates as shapefile

            :param crs:
            :param path:
            :return:
            """
            log.info(f"Saving instance of {self.__class__.__name__} as shapefile to {path}")
            self.to_geodf(crs).to_file(path, index=True)

    def _compute_labels(self, x: np.ndarray) -> np.ndarray:
        validate_coordinates(x)
        return self.clusterer._compute_labels(x)

    def fit(self, coordinates: TCoordinates):
        """
        Fitting to coordinates from a numpy array, a MultiPoint object or a GeoDataFrame with one Point per row

        :param coordinates:
        :return:
        """
        coordinates = extract_coordinates_array(coordinates)
        super().fit(coordinates)

    @timed
    def to_geodf(self, condition: Callable[[Cluster], bool] = None, crs='epsg:3857',
            include_noise=False) -> gp.GeoDataFrame:
        """
        GeoDataFrame containing all clusters found by the model.
        It is a concatenation of GeoDataFrames of individual clusters

        :param condition: if provided, only clusters fulfilling the condition will be included
        :param crs: projection. By default pseudo-mercator
        :param include_noise:
        :return: GeoDataFrame with all clusters indexed by their identifier
        """
        geodf = gp.GeoDataFrame()
        geodf.crs = crs
        for cluster in self.clusters(condition):
            geodf = pd.concat((geodf, cluster.to_geodf(crs=crs)))
        if include_noise:
            geodf = pd.concat((geodf, self.noise_cluster().to_geodf(crs=crs)))
        return geodf

    def plot(self, include_noise=False, condition=None, **kwargs):
        """
        Plots the resulting clusters with random coloring

        :param include_noise: Whether to include the noise cluster
        :param condition: If provided, only clusters fulfilling this condition will be included
        :param kwargs: passed to GeoDataFrame.plot
        :return:
        """
        geodf = self.to_geodf(condition=condition, include_noise=include_noise)
        geodf["color"] = np.random.random(len(geodf))
        if include_noise:
            geodf.loc[self.noiseLabel, "color"] = 0
        geodf.plot(column="color", **kwargs)

    # the overriding of the following methods is only necessary for getting the type annotations right
    # if mypy ever permits annotating nested classes correctly, these methods can be removed
    def get_cluster(self, cluster_id: int) -> Cluster:
        return super().get_cluster(cluster_id)

    def noise_cluster(self) -> Cluster:
        return super().noise_cluster()

    def clusters(self, condition: Callable[[Cluster], bool] = None) -> Iterable[Cluster]:
        return super().clusters(condition=condition)


class SkLearnCoordinateClustering(CoordinateEuclideanClusterer):
    """
    Wrapper around a sklearn clusterer. This class adds additional features like relabelling and convenient methods
    for handling geospatial data

    :param clusterer: a clusterer object compatible the sklearn API
    :param noise_label: label that is associated with the noise cluster or None
    :param min_cluster_size: if not None, clusters below this size will be labeled as noise
    :param max_cluster_size: if not None, clusters above this size will be labeled as noise
    """
    def __init__(self, clusterer: SkLearnClustererProtocol, noise_label=-1,
                 min_cluster_size: int = None, max_cluster_size: int = None):
        clusterer = SkLearnEuclideanClusterer(clusterer, noise_label=noise_label,
                           min_cluster_size=min_cluster_size, max_cluster_size=max_cluster_size)
        super().__init__(clusterer)



"""
This module contains various helper functions.
"""
import math
from typing import Any, Sequence, Union, TypeVar, List

T = TypeVar("T")


def count_none(*args: Any) -> int:
    """
    Counts the number of arguments that are None

    :param args: various arguments
    :return: the number of arguments that are None
    """
    c = 0
    for a in args:
        if a is None:
            c += 1
    return c


def count_not_none(*args: Any) -> int:
    """
    Counts the number of arguments that are not None

    :param args: various arguments
    :return: the number of arguments that are not None
    """
    return len(args) - count_none(*args)


def any_none(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if any of the arguments are None, False otherwise
    """
    return count_none(*args) > 0


def all_none(*args: Any) -> bool:
    """
    :param args: various arguments
    :return: True if all of the arguments are None, False otherwise
    """
    return count_none(*args) == len(args)


def check_not_nan_dict(d: dict):
    """
    Raises ValueError if any of the values in the given dictionary are NaN, reporting the respective keys

    :param d: a dictionary mapping to floats that are to be checked for NaN
    """
    invalid_keys = [k for k, v in d.items() if math.isnan(v)]
    if len(invalid_keys) > 0:
        raise ValueError(f"Got one or more NaN values: {invalid_keys}")


# noinspection PyUnusedLocal
def mark_used(*args):
    """
    Utility function to mark identifiers as used.
    The function does nothing.

    :param args: pass identifiers that shall be marked as used here
    """
    pass


def flatten_arguments(args: Sequence[Union[T, Sequence[T]]]) -> List[T]:
    """
    Main use case is to support both interfaces of the type f(T1, T2, ...) and f([T1, T2, ...]) simultaneously.
    It is assumed that if the latter form is passed, the arguments are either in a list or a tuple. Moreover,
    T cannot be a tuple or a list itself.

    Overall this function is not all too safe and one should be aware of what one is doing when using it
    """
    result = []
    for arg in args:
        if isinstance(arg, list) or isinstance(arg, tuple):
            result.extend(arg)
        else:
            result.append(arg)
    return result


import functools
import logging
import re
import sys
import types
from abc import ABC, abstractmethod
from typing import Union, List, Dict, Any, Sequence, Iterable, Optional, Mapping, Callable

reCommaWhitespacePotentiallyBreaks = re.compile(r",\s+")

log = logging.getLogger(__name__)


class StringConverter(ABC):
    """
    Abstraction for a string conversion mechanism
    """
    @abstractmethod
    def to_string(self, x) -> str:
        pass


def dict_string(d: Mapping, brackets: Optional[str] = None, converter: StringConverter = None):
    """
    Converts a dictionary to a string of the form "<key>=<value>, <key>=<value>, ...", optionally enclosed
    by brackets

    :param d: the dictionary
    :param brackets: a two-character string containing the opening and closing bracket to use, e.g. ``"{}"``;
        if None, do not use enclosing brackets
    :param converter: the string converter to use for values
    :return: the string representation
    """
    s = ', '.join([f'{k}={to_string(v, converter=converter, context=k)}' for k, v in d.items()])
    if brackets is not None:
        return brackets[:1] + s + brackets[-1:]
    else:
        return s


def list_string(l: Iterable[Any], brackets="[]", quote: Optional[str] = None, converter: StringConverter = None):
    """
    Converts a list or any other iterable to a string of the form "[<value>, <value>, ...]", optionally enclosed
    by different brackets or with the values quoted.

    :param l: the list
    :param brackets: a two-character string containing the opening and closing bracket to use, e.g. ``"[]"``;
        if None, do not use enclosing brackets
    :param quote: a 1-character string defining the quote to use around each value, e.g. ``"'"``.
    :param converter: the string converter to use for values
    :return: the string representation
    """
    def item(x):
        x = to_string(x, converter=converter, context="list")
        if quote is not None:
            return quote + x + quote
        else:
            return x
    s = ", ".join((item(x) for x in l))
    if brackets is not None:
        return brackets[:1] + s + brackets[-1:]
    else:
        return s


def to_string(x, converter: StringConverter = None, apply_converter_to_non_complex_objects=True, context=None):
    """
    Converts the given object to a string, with proper handling of lists, tuples and dictionaries, optionally using a converter.
    The conversion also removes unwanted line breaks (as present, in particular, in sklearn's string representations).

    :param x: the object to convert
    :param converter: the converter with which to convert objects to strings
    :param apply_converter_to_non_complex_objects: whether to apply/pass on the converter (if any) not only when converting complex objects
        but also non-complex, primitive objects; use of this flag enables converters to implement their conversion functionality using this
        function for complex objects without causing an infinite recursion.
    :param context: context in which the object is being converted (e.g. dictionary key for case where x is the corresponding
        dictionary value), only for debugging purposes (will be reported in log messages upon recursion exception)
    :return: the string representation
    """
    try:
        if type(x) == list:
            return list_string(x, converter=converter)
        elif type(x) == tuple:
            return list_string(x, brackets="()", converter=converter)
        elif type(x) == dict:
            return dict_string(x, brackets="{}", converter=converter)
        elif type(x) == types.MethodType:
            # could be bound method of a ToStringMixin instance (which would print the repr of the instance, which can potentially cause
            # an infinite recursion)
            return f"Method[{x.__name__}]"
        else:
            if converter and apply_converter_to_non_complex_objects:
                s = converter.to_string(x)
            else:
                s = str(x)

            # remove any unwanted line breaks and indentation after commas (as generated, for example, by sklearn objects)
            s = reCommaWhitespacePotentiallyBreaks.sub(", ", s)

            return s
    except RecursionError as e:
        log.error(f"Recursion in string conversion detected; context={context}")
        raise


def object_repr(obj, member_names_or_dict: Union[List[str], Dict[str, Any]]):
    if type(member_names_or_dict) == dict:
        members_dict = member_names_or_dict
    else:
        members_dict = {m: to_string(getattr(obj, m)) for m in member_names_or_dict}
    return f"{obj.__class__.__name__}[{dict_string(members_dict)}]"


def or_regex_group(allowed_names: Sequence[str]):
    """

    :param allowed_names: strings to include as literals in the regex
    :return: a regular expression string of the form (<name1>| ...|<nameN>), which any of the given names
    """
    allowed_names = [re.escape(name) for name in allowed_names]
    return r"(%s)" % "|".join(allowed_names)


def function_name(x: Callable) -> str:
    if isinstance(x, functools.partial):
        return function_name(x.func)
    elif hasattr(x, "__name__"):
        return x.__name__
    else:
        return str(x)


class ToStringMixin:
    """
    Provides implementations for ``__str__`` and ``__repr__`` which are based on the format ``"<class name>[<object info>]"`` and
    ``"<class name>[id=<object id>, <object info>]"`` respectively, where ``<object info>`` is usually a list of entries of the
    form ``"<name>=<value>, ..."``.

    By default, ``<class name>`` will be the qualified name of the class, and ``<object info>`` will include all properties
    of the class, including private ones starting with an underscore (though the underscore will be dropped in the string
    representation).

        * To exclude private properties, override :meth:`_toStringExcludePrivate` to return True. If there are exceptions
          (and some private properties shall be retained), additionally override :meth:`_toStringExcludeExceptions`.
        * To exclude a particular set of properties, override :meth:`_toStringExcludes`.
        * To include only select properties (introducing inclusion semantics), override :meth:`_toStringIncludes`.
        * To add values to the properties list that aren't actually properties of the object (i.e. derived properties),
          override :meth:`_toStringAdditionalEntries`.
        * To define a fully custom representation for ``<object info>`` which is not based on the above principles, override
          :meth:`_toStringObjectInfo`.

    For well-defined string conversions within a class hierarchy, it can be a good practice to define additional
    inclusions/exclusions by overriding the respective method once more and basing the return value on an extended
    version of the value returned by superclass.
    In some cases, the requirements of a subclass can be at odds with the definitions in the superclass: The superclass
    may make use of exclusion semantics, but the subclass may want to use inclusion semantics (and include
    only some of the many properties it adds). In this case, if the subclass used :meth:`_toStringInclude`, the exclusion semantics
    of the superclass would be void and none of its properties would actually be included.
    In such cases, override :meth:`_toStringIncludesForced` to add inclusions regardless of the semantics otherwise used along
    the class hierarchy.

    .. document private functions
    .. automethod:: _tostring_class_name
    .. automethod:: _tostring_object_info
    .. automethod:: _tostring_excludes
    .. automethod:: _tostring_exclude_exceptions
    .. automethod:: _tostring_includes
    .. automethod:: _tostring_includes_forced
    .. automethod:: _tostring_additional_entries
    .. automethod:: _tostring_exclude_private
    """
    _TOSTRING_INCLUDE_ALL = "__all__"

    def _tostring_class_name(self):
        """
        :return: the string use for <class name> in the string representation ``"<class name>[<object info]"``
        """
        return type(self).__qualname__

    def _tostring_properties(self,
            exclude: Optional[Union[str, Iterable[str]]] = None,
            include: Optional[Union[str, Iterable[str]]] = None,
            exclude_exceptions: Optional[List[str]] = None,
            include_forced: Optional[List[str]] = None,
            additional_entries: Dict[str, Any] = None,
            converter: StringConverter = None) -> str:
        """
        Creates a string of the class attributes, with optional exclusions/inclusions/additions.
        Exclusions take precedence over inclusions.

        :param exclude: attributes to be excluded
        :param include: attributes to be included; if non-empty, only the specified attributes will be printed (bar the ones
            excluded by ``exclude``)
        :param include_forced: additional attributes to be included
        :param additional_entries: additional key-value entries to be added
        :param converter: the string converter to use; if None, use default (which avoids infinite recursions)
        :return: a string containing entry/property names and values
        """
        def mklist(x):
            if x is None:
                return []
            if type(x) == str:
                return [x]
            return x

        exclude = mklist(exclude)
        include = mklist(include)
        include_forced = mklist(include_forced)
        exclude_exceptions = mklist(exclude_exceptions)

        def is_excluded(k):
            if k in include_forced or k in exclude_exceptions:
                return False
            if k in exclude:
                return True
            if self._tostring_exclude_private():
                is_private = k.startswith("_")
                return is_private
            else:
                return False

        # determine relevant attribute dictionary
        if len(include) == 1 and include[0] == self._TOSTRING_INCLUDE_ALL:  # exclude semantics (include everything by default)
            attribute_dict = self.__dict__
        else:  # include semantics (include only inclusions)
            attribute_dict = {k: getattr(self, k) for k in set(include + include_forced)
                if hasattr(self, k) and k != self._TOSTRING_INCLUDE_ALL}

        # apply exclusions and remove underscores from attribute names
        d = {k.strip("_"): v for k, v in attribute_dict.items() if not is_excluded(k)}

        if additional_entries is not None:
            d.update(additional_entries)

        if converter is None:
            converter = self._StringConverterAvoidToStringMixinRecursion(self)
        return dict_string(d, converter=converter)

    def _tostring_object_info(self) -> str:
        """
        Override this method to use a fully custom definition of the ``<object info>`` part in the full string
        representation ``"<class name>[<object info>]"`` to be generated.
        As soon as this method is overridden, any property-based exclusions, inclusions, etc. will have no effect
        (unless the implementation is specifically designed to make use of them - as is the default
        implementation).
        NOTE: Overrides must not internally use super() because of a technical limitation in the proxy
        object that is used for nested object structures.

        :return: a string containing the string to use for ``<object info>``
        """
        return self._tostring_properties(exclude=self._tostring_excludes(), include=self._tostring_includes(),
            exclude_exceptions=self._tostring_exclude_exceptions(), include_forced=self._tostring_includes_forced(),
            additional_entries=self._tostring_additional_entries())

    def _tostring_excludes(self) -> List[str]:
        """
        Makes the string representation exclude the returned attributes.
        This method can be conveniently overridden by subclasses which can call super and extend the list returned.

        This method will only have no effect if :meth:`_toStringObjectInfo` is overridden to not use its result.

        :return: a list of attribute names
        """
        return []

    def _tostring_includes(self) -> List[str]:
        """
        Makes the string representation include only the returned attributes (i.e. introduces inclusion semantics);
        By default, the list contains only a marker element, which is interpreted as "all attributes included".

        This method can be conveniently overridden by sub-classes which can call super and extend the list returned.
        Note that it is not a problem for a list containing the aforementioned marker element (which stands for all attributes)
        to be extended; the marker element will be ignored and only the user-added elements will be considered as included.

        Note: To add an included attribute in a sub-class, regardless of any super-classes using exclusion or inclusion semantics,
        use _toStringIncludesForced instead.

        This method will have no effect if :meth:`_toStringObjectInfo` is overridden to not use its result.

        :return: a list of attribute names to be included in the string representation
        """
        return [self._TOSTRING_INCLUDE_ALL]

    # noinspection PyMethodMayBeStatic
    def _tostring_includes_forced(self) -> List[str]:
        """
        Defines a list of attribute names that are required to be present in the string representation, regardless of the
        instance using include semantics or exclude semantics, thus facilitating added inclusions in sub-classes.

        This method will have no effect if :meth:`_toStringObjectInfo` is overridden to not use its result.

        :return: a list of attribute names
        """
        return []

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        """
        :return: a dictionary of entries to be included in the ``<object info>`` part of the string representation
        """
        return {}

    def _tostring_exclude_private(self) -> bool:
        """
        :return: whether to exclude properties that are private (start with an underscore); explicitly included attributes
            will still be considered - as will properties exempt from the rule via :meth:`toStringExcludeException`.
        """
        return False

    def _tostring_exclude_exceptions(self) -> List[str]:
        """
        Defines attribute names which should not be excluded even though other rules (particularly the exclusion of private members
        via :meth:`_toStringExcludePrivate`) would otherwise exclude them.

        :return: a list of attribute names
        """
        return []

    def __str__(self):
        return f"{self._tostring_class_name()}[{self._tostring_object_info()}]"

    def __repr__(self):
        info = f"id={id(self)}"
        property_info = self._tostring_object_info()
        if len(property_info) > 0:
            info += ", " + property_info
        return f"{self._tostring_class_name()}[{info}]"

    def pprint(self, file=sys.stdout):
        """
        Prints a prettily formatted string representation of the object (with line breaks and indentations)
        to ``stdout`` or the given file.

        :param file: the file to print to
        """
        print(self.pprints(), file=file)

    def pprints(self) -> str:
        """
        :return: a prettily formatted string representation with line breaks and indentations
        """
        return pretty_string_repr(self)

    class _StringConverterAvoidToStringMixinRecursion(StringConverter):
        """
        Avoids recursions when converting objects implementing :class:`ToStringMixin` which may contain themselves to strings.
        Use of this object prevents infinite recursions caused by a :class:`ToStringMixin` instance recursively containing itself in
        either a property of another :class:`ToStringMixin`, a list or a tuple.
        It handles all :class:`ToStringMixin` instances recursively encountered.

        A previously handled instance is converted to a string of the form "<class name>[<<]".
        """
        def __init__(self, *handled_objects: "ToStringMixin"):
            """
            :param handled_objects: objects which are initially assumed to have been handled already
            """
            self._handled_to_string_mixin_ids = set([id(o) for o in handled_objects])

        def to_string(self, x) -> str:
            if isinstance(x, ToStringMixin):
                oid = id(x)
                if oid in self._handled_to_string_mixin_ids:
                    return f"{x._tostring_class_name()}[<<]"
                self._handled_to_string_mixin_ids.add(oid)
                return str(self._ToStringMixinProxy(x, self))
            else:
                return to_string(x, converter=self, apply_converter_to_non_complex_objects=False, context=x.__class__)

        class _ToStringMixinProxy:
            """
            A proxy object which wraps a ToStringMixin to ensure that the converter is applied when creating the properties string.
            The proxy is to achieve that all ToStringMixin methods that aren't explicitly overwritten are bound to this proxy
            (rather than the original object), such that the transitive call to _toStringProperties will call the new
            implementation.
            """

            # methods where we assume that they could transitively call _toStringProperties (others are assumed not to)
            TOSTRING_METHODS_TRANSITIVELY_CALLING_TOSTRINGPROPERTIES = {"_tostring_object_info"}

            def __init__(self, x: "ToStringMixin", converter):
                self.x = x
                self.converter = converter

            def _tostring_properties(self, *args, **kwargs):
                return self.x._tostring_properties(*args, **kwargs, converter=self.converter)

            def _tostring_class_name(self):
                return self.x._tostring_class_name()

            def __getattr__(self, attr: str):
                if attr.startswith("_tostring"):  # ToStringMixin method which we may bind to use this proxy to ensure correct transitive call
                    method = getattr(self.x.__class__, attr)
                    obj = self if attr in self.TOSTRING_METHODS_TRANSITIVELY_CALLING_TOSTRINGPROPERTIES else self.x
                    return lambda *args, **kwargs: method(obj, *args, **kwargs)
                else:
                    return getattr(self.x, attr)

            def __str__(self: "ToStringMixin"):
                return ToStringMixin.__str__(self)


def pretty_string_repr(s: Any, initial_indentation_level=0, indentation_string="    "):
    """
    Creates a pretty string representation (using indentations) from the given object/string representation (as generated, for example, via
    ToStringMixin). An indentation level is added for every opening bracket.

    :param s: an object or object string representation
    :param initial_indentation_level: the initial indentation level
    :param indentation_string: the string which corresponds to a single indentation level
    :return: a reformatted version of the input string with added indentations and line breaks
    """
    if type(s) != str:
        s = str(s)
    indent = initial_indentation_level
    result = indentation_string * indent
    i = 0

    def nl():
        nonlocal result
        result += "\n" + (indentation_string * indent)

    def take(cnt=1):
        nonlocal result, i
        result += s[i:i+cnt]
        i += cnt

    def find_matching(j):
        start = j
        op = s[j]
        cl = {"[": "]", "(": ")", "'": "'"}[s[j]]
        is_bracket = cl != s[j]
        stack = 0
        while j < len(s):
            if s[j] == op and (is_bracket or j == start):
                stack += 1
            elif s[j] == cl:
                stack -= 1
            if stack == 0:
                return j
            j += 1
        return None

    brackets = "[("
    quotes = "'"
    while i < len(s):
        is_bracket = s[i] in brackets
        is_quote = s[i] in quotes
        if is_bracket or is_quote:
            i_match = find_matching(i)
            take_full_match_without_break = False
            if i_match is not None:
                k = i_match + 1
                full_match = s[i:k]
                take_full_match_without_break = is_quote or not("=" in full_match and "," in full_match)
                if take_full_match_without_break:
                    take(k-i)
            if not take_full_match_without_break:
                take(1)
                indent += 1
                nl()
        elif s[i] in "])":
            take(1)
            indent -= 1
        elif s[i:i+2] == ", ":
            take(2)
            nl()
        else:
            take(1)

    return result


class TagBuilder:
    """
    Assists in building strings made up of components that are joined via a glue string
    """
    def __init__(self, *initial_components: str, glue="_"):
        """
        :param initial_components: initial components to always include at the beginning
        :param glue: the glue string which joins components
        """
        self.glue = glue
        self.components = list(initial_components)

    def with_component(self, component: str):
        self.components.append(component)

    def with_conditional(self, cond: bool, component: str):
        """
        Conditionally adds the given component

        :param cond: the condition
        :param component: the component to add if the condition holds
        :return: the builder
        """
        if cond:
            self.components.append(component)
        return self

    def with_alternative(self, cond: bool, true_component: str, false_component: str):
        """
        Adds a component depending on a condition

        :param cond: the condition
        :param true_component: the component to add if the condition holds
        :param false_component: the component to add if the condition does not hold
        :return: the builder
        """
        self.components.append(true_component if cond else false_component)
        return self

    def build(self):
        """
        :return: the string (with all components joined)
        """
        return self.glue.join(self.components)



import logging
from abc import ABC, abstractmethod
import math
from typing import Tuple, Sequence, Optional, Union, List, Iterator

import numpy as np
import pandas as pd
import sklearn.preprocessing
import torch
from torch.autograd import Variable

from .. import normalisation
from ..data import DataFrameSplitter, DataFrameSplitterFractional
from ..data_transformation import DFTSkLearnTransformer
from ..util.dtype import to_float_array
from ..util.pickle import setstate


log = logging.getLogger(__name__)


def to_tensor(d: Union[torch.Tensor, np.ndarray, list], cuda=False):
    if not isinstance(d, torch.Tensor):
        if isinstance(d, np.ndarray):
            d = torch.from_numpy(d)
        elif isinstance(d, list):
            d = torch.from_numpy(np.array(d))
        else:
            raise ValueError()
    if cuda:
        d.cuda()
    return d


class TensorScaler(ABC):
    @abstractmethod
    def cuda(self):
        """
        Makes this scaler's components use CUDA
        """
        pass

    @abstractmethod
    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies scaling/normalisation to the given tensor
        :param tensor: the tensor to scale/normalise
        :return: the scaled/normalised tensor
        """
        pass

    @abstractmethod
    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Applies the inverse of method normalise to the given tensor
        :param tensor: the tensor to denormalise
        :return: the denormalised tensor
        """
        pass


class TensorScalerCentreAndScale(TensorScaler):
    def __init__(self, centre: Optional[torch.Tensor] = None, scale: Optional[torch.Tensor] = None):
        self.centre = centre
        self.scale = scale

    def cuda(self):
        if self.scale is not None:
            self.scale = self.scale.cuda()
        if self.centre is not None:
            self.centre = self.centre.cuda()

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.centre is not None:
            tensor -= self.centre
        if self.scale is not None:
            tensor *= self.scale
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        if self.scale is not None:
            tensor /= self.scale
        if self.centre is not None:
            tensor += self.centre
        return tensor


class TensorScalerFromVectorDataScaler(TensorScalerCentreAndScale):
    def __init__(self, vector_data_scaler: normalisation.VectorDataScaler, cuda: bool):
        if vector_data_scaler.scale is not None:
            inv_scale = torch.from_numpy(vector_data_scaler.scale).float()
            scale = 1.0 / inv_scale
        else:
            scale = None
        centre = vector_data_scaler.translate
        if centre is not None:
            centre = torch.from_numpy(vector_data_scaler.translate).float()
        super().__init__(centre=centre, scale=scale)
        if cuda:
            self.cuda()

    def __setstate__(self, state):
        if "translate" in state:
            if state["scale"] is not None:  # old representation where scale is actually inverse scale
                state["scale"] = 1.0 / state["scale"]
        setstate(TensorScalerFromVectorDataScaler, self, state, renamed_properties={"translate": "centre"})


class TensorScalerIdentity(TensorScaler):
    def cuda(self):
        pass

    def normalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def denormalise(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor


class TensorScalerFromDFTSkLearnTransformer(TensorScalerCentreAndScale):
    def __init__(self, dft: DFTSkLearnTransformer):
        trans = dft.sklearnTransformer
        if isinstance(trans, sklearn.preprocessing.RobustScaler):
            centre = trans.center_
            scale = trans.scale_
            is_reciprocal_scale = True
        else:
            raise ValueError(f"sklearn transformer of type '{trans.__class__}' is unhandled")
        if centre is not None:
            centre = torch.from_numpy(centre).float()
        if scale is not None:
            scale = torch.from_numpy(scale).float()
            if is_reciprocal_scale:
                scale = 1.0 / scale
        super().__init__(centre=centre, scale=scale)


class Tensoriser(ABC):
    """
    Represents a method for transforming a data frame into one or more tensors to be processed by a neural network model
    """
    def tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
        result = self._tensorise(df)
        if type(result) == list:
            lengths = set(map(len, result))
            if len(lengths) != 1:
                raise Exception("Lengths of tensors inconsistent")
            length = lengths.pop()
        else:
            length = len(result)
        if length != len(df):
            raise Exception(f"{self} produced result of length {length} for DataFrame of shape {df.shape}")
        return result

    @abstractmethod
    def _tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
        pass

    @abstractmethod
    def fit(self, df: pd.DataFrame, model=None):
        """
        :param df: the data frame with which to fit this tensoriser
        :param model: the model in the context of which the fitting takes place (if any).
            The fitting process may set parameters within the model that can only be determined from the (pre-tensorised) data.
        """
        pass


class RuleBasedTensoriser(Tensoriser, ABC):
    """
    Base class for tensorisers which transform data frames into tensors based on a predefined set of rules and do not require fitting
    """
    def fit(self, df: pd.DataFrame, model=None):
        pass


class TensoriserDataFrameFloatValuesMatrix(RuleBasedTensoriser):
    def _tensorise(self, df: pd.DataFrame) -> np.ndarray:
        return torch.from_numpy(to_float_array(df)).float()


class TensoriserClassLabelIndices(RuleBasedTensoriser):
    def _tensorise(self, df: pd.DataFrame) -> np.ndarray:
        if len(df.columns) != 1:
            raise ValueError("Expected a single column containing the class label indices")
        return torch.from_numpy(df[df.columns[0]].values).long()


class DataUtil(ABC):
    """Interface for DataUtil classes, which are used to process data for neural networks"""

    @abstractmethod
    def split_into_tensors(self, fractional_size_of_first_set) \
            -> Tuple[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        """
        Splits the data set

        :param fractional_size_of_first_set: the desired fractional size in
        :return: a tuple (A, B) where A and B are tuples (in, out) with input and output data
        """
        pass

    @abstractmethod
    def get_output_tensor_scaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model outputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def get_input_tensor_scaler(self) -> TensorScaler:
        """
        Gets the scaler with which to scale model inputs

        :return: the scaler
        """
        pass

    @abstractmethod
    def model_output_dim(self) -> int:
        """
        :return: the dimensionality that is to be output by the model to be trained
        """
        pass

    @abstractmethod
    def input_dim(self):
        """
        :return: the dimensionality of the input or None if it is variable
        """
        pass


class VectorDataUtil(DataUtil):
    def __init__(self,
            inputs: pd.DataFrame,
            outputs: pd.DataFrame,
            cuda: bool,
            normalisation_mode=normalisation.NormalisationMode.NONE,
            differing_output_normalisation_mode=None,
            input_tensoriser: Optional[Tensoriser] = None,
            output_tensoriser: Optional[Tensoriser] = None,
            data_frame_splitter: Optional[DataFrameSplitter] = None):
        """
        :param inputs: the data frame of inputs
        :param outputs: the data frame of outputs
        :param cuda: whether to apply CUDA
        :param normalisation_mode: the normalisation mode to use for inputs and (unless differingOutputNormalisationMode is specified)
            outputs
        :param differing_output_normalisation_mode: the normalisation mode to apply to outputs, overriding normalisationMode;
            if None, use normalisationMode
        """
        if inputs.shape[0] != outputs.shape[0]:
            raise ValueError("Output length must be equal to input length")
        self.inputs = inputs
        self.outputs = outputs
        self.inputTensoriser = input_tensoriser if input_tensoriser is not None else TensoriserDataFrameFloatValuesMatrix()
        self.outputTensoriser = output_tensoriser if output_tensoriser is not None else TensoriserDataFrameFloatValuesMatrix()
        self.inputVectorDataScaler = normalisation.VectorDataScaler(self.inputs, normalisation_mode)
        self.inputTensorScaler = TensorScalerFromVectorDataScaler(self.inputVectorDataScaler, cuda)
        self.outputVectorDataScaler = normalisation.VectorDataScaler(self.outputs,
            normalisation_mode if differing_output_normalisation_mode is None else differing_output_normalisation_mode)
        self.outputTensorScaler = TensorScalerFromVectorDataScaler(self.outputVectorDataScaler, cuda)
        self.dataFrameSplitter = data_frame_splitter

    def __len__(self):
        return len(self.inputs)

    def get_output_tensor_scaler(self):
        return self.outputTensorScaler

    def get_input_tensor_scaler(self):
        return self.inputTensorScaler

    def _compute_split_indices(self, fractional_size_of_first_set):
        splitter = self.dataFrameSplitter
        if splitter is None:
            # By default, we use a simple fractional split without shuffling.
            # Shuffling is usually unnecessary, because in evaluation contexts, the data may have already been shuffled by the evaluator
            # (unless explicitly disabled by the user). Furthermore, not shuffling gives the user the possibility to manually
            # order the data in ways that result in desirable fractional splits (though the user may, of course, simply override
            # the splitter to achieve any desired split).
            splitter = DataFrameSplitterFractional(shuffle=False)
        indices_a, indices_b = splitter.compute_split_indices(self.inputs, fractional_size_of_first_set)
        return indices_a, indices_b

    def split_into_tensors(self, fractional_size_of_first_set):
        indices_a, indices_b = self._compute_split_indices(fractional_size_of_first_set)
        a = self._tensors_for_indices(indices_a)
        b = self._tensors_for_indices(indices_b)
        return a, b

    def _data_frames_for_indices(self, indices):
        input_df = self.inputs.iloc[indices]
        output_df = self.outputs.iloc[indices]
        return input_df, output_df

    def _tensors_for_indices(self, indices):
        input_df, output_df = self._data_frames_for_indices(indices)
        return self._tensors_for_data_frames(input_df, output_df)

    def _tensors_for_data_frames(self, input_df, output_df):
        # apply normalisation (if any)
        if self.inputVectorDataScaler.normalisation_mode != normalisation.NormalisationMode.NONE:
            input_df = pd.DataFrame(self.inputVectorDataScaler.get_normalised_array(input_df), columns=input_df.columns,
                index=input_df.index)
        if self.outputVectorDataScaler.normalisation_mode != normalisation.NormalisationMode.NONE:
            output_df = pd.DataFrame(self.outputVectorDataScaler.get_normalised_array(output_df), columns=output_df.columns,
                index=output_df.index)

        return self.inputTensoriser.tensorise(input_df), self.outputTensoriser.tensorise(output_df)

    def split_into_data_sets(self, fractional_size_of_first_set, cuda: bool, tensorise_dynamically=False) \
            -> Tuple["TorchDataSet", "TorchDataSet"]:
        if not tensorise_dynamically:
            (xA, yA), (xB, yB) = self.split_into_tensors(fractional_size_of_first_set)
            return TorchDataSetFromTensors(xA, yA, cuda), TorchDataSetFromTensors(xB, yB, cuda)
        else:
            if self.inputVectorDataScaler.normalisation_mode != normalisation.NormalisationMode.NONE or \
                    self.outputVectorDataScaler.normalisation_mode != normalisation.NormalisationMode.NONE:
                raise Exception("Dynamic tensorisation is not supported when using data scaling")
            indices_a, indices_b = self._compute_split_indices(fractional_size_of_first_set)
            input_a, output_a = self._data_frames_for_indices(indices_a)
            input_b, output_b = self._data_frames_for_indices(indices_b)
            ds_a = TorchDataSetFromDataFramesDynamicallyTensorised(input_a, output_a, cuda, input_tensoriser=self.inputTensoriser,
                output_tensoriser=self.outputTensoriser)
            ds_b = TorchDataSetFromDataFramesDynamicallyTensorised(input_b, output_b, cuda, input_tensoriser=self.inputTensoriser,
                output_tensoriser=self.outputTensoriser)
            return ds_a, ds_b

    def input_dim(self):
        return self.inputs.shape[1]

    def output_dim(self):
        """
        :return: the dimensionality of the outputs (ground truth values)
        """
        return self.outputs.shape[1]

    def model_output_dim(self):
        return self.output_dim()


class ClassificationVectorDataUtil(VectorDataUtil):
    def __init__(self,
            inputs: pd.DataFrame,
            outputs: pd.DataFrame,
            cuda,
            num_classes,
            normalisation_mode=normalisation.NormalisationMode.NONE,
            input_tensoriser: Tensoriser = None,
            output_tensoriser: Tensoriser = None,
            data_frame_splitter: Optional[DataFrameSplitter] = None):
        if len(outputs.columns) != 1:
            raise Exception(f"Exactly one output dimension (the class index) is required, got {len(outputs.columns)}")
        super().__init__(inputs, outputs, cuda, normalisation_mode=normalisation_mode,
            differing_output_normalisation_mode=normalisation.NormalisationMode.NONE, input_tensoriser=input_tensoriser,
            output_tensoriser=TensoriserClassLabelIndices() if output_tensoriser is None else output_tensoriser,
            data_frame_splitter=data_frame_splitter)
        self.numClasses = num_classes

    def model_output_dim(self):
        return self.numClasses


class TorchDataSet:
    @abstractmethod
    def iter_batches(self, batch_size: int, shuffle: bool = False, input_only=False) -> Iterator[Union[Tuple[torch.Tensor, torch.Tensor],
            Tuple[Sequence[torch.Tensor], torch.Tensor], torch.Tensor, Sequence[torch.Tensor]]]:
        """
        Provides an iterator over batches from the data set.

        :param batch_size: the maximum size of each batch
        :param shuffle: whether to shuffle the data set
        :param input_only: whether to provide only inputs (rather than inputs and corresponding outputs).
            If true, provide only inputs, where inputs can either be a tensor or a tuple of tensors.
            If false, provide a pair (i, o) with inputs and corresponding outputs (o is always a tensor).
            Some data sets may only be able to provide inputs, in which case inputOnly=False should lead to an
            exception.
        """
        pass

    @abstractmethod
    def size(self) -> Optional[int]:
        """
        Returns the total size of the data set (number of data points) if it is known.

        :return: the number of data points or None of the size is not known.
        """
        pass


class TorchDataSetProvider:
    def __init__(self, input_tensor_scaler: Optional[TensorScaler] = None, output_tensor_scaler: Optional[TensorScaler] = None,
            input_dim: Optional[int] = None, model_output_dim: int = None):
        if input_tensor_scaler is None:
            input_tensor_scaler = TensorScalerIdentity()
        if output_tensor_scaler is None:
            output_tensor_scaler = TensorScalerIdentity()
        if model_output_dim is None:
            raise ValueError("The model output dimension must be provided")
        self.inputTensorScaler = input_tensor_scaler
        self.outputTensorScaler = output_tensor_scaler
        self.inputDim = input_dim
        self.modelOutputDim = model_output_dim

    @abstractmethod
    def provide_split(self, fractional_size_of_first_set: float) -> Tuple[TorchDataSet, TorchDataSet]:
        """
        Provides two data sets, which could, for example, serve as training and validation sets.

        :param fractional_size_of_first_set: the fractional size of the first data set
        :return: a tuple of data sets (A, B) where A has (approximately) the given fractional size and B encompasses
            the remainder of the data
        """
        pass

    def get_output_tensor_scaler(self) -> TensorScaler:
        return self.outputTensorScaler

    def get_input_tensor_scaler(self) -> TensorScaler:
        return self.inputTensorScaler

    def get_model_output_dim(self) -> int:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
        """
        return self.modelOutputDim

    def get_input_dim(self) -> Optional[int]:
        """
        :return: the number of output dimensions that would be required to be generated by the model to match this dataset.
            For models that accept variable input sizes (such as RNNs), this may be None.
        """
        return self.inputDim


class TensorTuple:
    """
    Represents a tuple of tensors (or a single tensor) and can be used to manipulate the contained tensors simultaneously
    """
    def __init__(self, tensors: Union[torch.Tensor, Sequence[torch.Tensor]]):
        if isinstance(tensors, torch.Tensor):
            tensors = [tensors]
        lengths = set(map(len, tensors))
        if len(lengths) != 1:
            raise ValueError("Not all tensors are of the same length")
        self.length = lengths.pop()
        self.tensors = tensors

    def __len__(self):
        return self.length

    def __getitem__(self, key) -> "TensorTuple":
        t = tuple((t[key] for t in self.tensors))
        return TensorTuple(t)

    def cuda(self) -> "TensorTuple":
        return TensorTuple([t.cuda() for t in self.tensors])

    def tuple(self) -> Sequence[torch.Tensor]:
        return tuple(self.tensors)

    def item(self) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
        if len(self.tensors) == 1:
            return self.tensors[0]
        else:
            return self.tuple()

    def concat(self, other: "TensorTuple") -> "TensorTuple":
        if len(self.tensors) != len(other.tensors):
            raise ValueError("Tensor tuples are incompatible")
        tensors = [torch.cat([a, b], dim=0) for a, b in zip(self.tensors, other.tensors)]
        return TensorTuple(tensors)


class TorchDataSetFromTensors(TorchDataSet):
    def __init__(self, x: Union[torch.Tensor, Sequence[torch.Tensor]], y: Optional[torch.Tensor], cuda: bool):
        """
        :param x: the input tensor(s); if more than one, they must be of the same length (and a slice of each shall be provided to the
            model as an input in each batch)
        :param y: the output tensor
        :param cuda: whether any generated tensors shall be moved to the selected CUDA device
        """
        x = TensorTuple(x)
        y = TensorTuple(y) if y is not None else None
        if y is not None and len(x) != len(y):
            raise ValueError("Tensors are not of the same length")
        self.x = x
        self.y = y
        self.cuda = cuda

    def iter_batches(self, batch_size: int, shuffle: bool = False, input_only=False) -> Iterator[Union[Tuple[torch.Tensor, torch.Tensor],
            Tuple[Sequence[torch.Tensor], torch.Tensor], torch.Tensor, Sequence[torch.Tensor]]]:
        tensor_tuples = (self.x, self.y) if not input_only and self.y is not None else (self.x,)
        yield from self._get_batches(tensor_tuples, batch_size, shuffle)

    def _get_batches(self, tensor_tuples: Sequence[TensorTuple], batch_size, shuffle):
        length = len(tensor_tuples[0])
        if shuffle:
            index = torch.randperm(length)
        else:
            index = torch.LongTensor(range(length))
        start_idx = 0
        while start_idx < length:
            remaining_items = length - start_idx
            is_second_last_batch = remaining_items <= 2*batch_size and remaining_items > batch_size
            if is_second_last_batch:
                # to avoid cases where the last batch is excessively small (1 item in the worst case, where e.g. batch
                # normalisation would not be applicable), we evenly distribute the items across the last two batches
                adjusted_batch_size = math.ceil(remaining_items / 2)
                end_idx = min(length, start_idx + adjusted_batch_size)
            else:
                end_idx = min(length, start_idx + batch_size)
            excerpt = index[start_idx:end_idx]
            batch = []
            for tensorTuple in tensor_tuples:
                if len(tensorTuple) != length:
                    raise Exception("Passed tensors of differing lengths")
                t = tensorTuple[excerpt]
                if self.cuda:
                    t = t.cuda()
                item = t.item()
                if type(item) == tuple:
                    item = tuple(Variable(t) for t in item)
                else:
                    item = Variable(item)
                batch.append(item)
            if len(batch) == 1:
                yield batch[0]
            else:
                yield tuple(batch)
            start_idx = end_idx

    def size(self):
        return len(self.x)


class TorchDataSetFromDataFramesPreTensorised(TorchDataSetFromTensors):
    def __init__(self, input_df: pd.DataFrame, output_df: Optional[pd.DataFrame], cuda: bool,
            input_tensoriser: Optional[Tensoriser] = None, output_tensoriser: Optional[Tensoriser] = None):
        if input_tensoriser is None:
            input_tensoriser = TensoriserDataFrameFloatValuesMatrix()
        log.debug(f"Applying {input_tensoriser} to data frame of length {len(input_df)} ...")
        input_tensors = input_tensoriser.tensorise(input_df)
        if output_df is not None:
            if output_tensoriser is None:
                output_tensoriser = TensoriserDataFrameFloatValuesMatrix()
            log.debug(f"Applying {output_tensoriser} to data frame of length {len(output_df)} ...")
            output_tensors = output_tensoriser.tensorise(output_df)
        else:
            output_tensors = None
        super().__init__(input_tensors, output_tensors, cuda)


class TorchDataSetFromDataFramesDynamicallyTensorised(TorchDataSet):
    def __init__(self, input_df: pd.DataFrame, output_df: Optional[pd.DataFrame], cuda: bool,
            input_tensoriser: Optional[Tensoriser] = None, output_tensoriser: Optional[Tensoriser] = None):
        self.inputDF = input_df
        self.outputDF = output_df
        self.cuda = cuda
        if input_tensoriser is None:
            input_tensoriser = TensoriserDataFrameFloatValuesMatrix()
        self.inputTensoriser = input_tensoriser
        if output_df is not None:
            if len(input_df) != len(output_df):
                raise ValueError("Lengths of input and output data frames must be equal")
            if output_tensoriser is None:
                output_tensoriser = TensoriserDataFrameFloatValuesMatrix()
        self.outputTensoriser = output_tensoriser

    def size(self) -> Optional[int]:
        return len(self.inputDF)

    def iter_batches(self, batch_size: int, shuffle: bool = False, input_only=False):
        length = len(self.inputDF)
        if shuffle:
            index = torch.randperm(length)
        else:
            index = torch.LongTensor(range(length))
        i = 0
        while i < length:
            batch_indices = index[i:i + batch_size]
            input_tensors = TensorTuple(self.inputTensoriser.tensorise(self.inputDF.iloc[batch_indices]))
            if self.cuda:
                input_tensors = input_tensors.cuda()
            if input_only:
                yield input_tensors.item()
            else:
                output_tensors = TensorTuple(self.outputTensoriser.tensorise(self.outputDF.iloc[batch_indices]))
                if self.cuda:
                    output_tensors = output_tensors.cuda()
                yield input_tensors.item(), output_tensors.item()
            i += batch_size


class TorchDataSetFromDataFrames(TorchDataSet):
    def __init__(self, input_df: pd.DataFrame, output_df: Optional[pd.DataFrame], cuda: bool,
            input_tensoriser: Optional[Tensoriser] = None, output_tensoriser: Optional[Tensoriser] = None,
            tensorise_dynamically=False):
        if tensorise_dynamically:
            self._torchDataSet: TorchDataSet = TorchDataSetFromDataFramesDynamicallyTensorised(input_df, output_df, cuda,
                input_tensoriser=input_tensoriser, output_tensoriser=output_tensoriser)
        else:
            self._torchDataSet: TorchDataSet = TorchDataSetFromDataFramesPreTensorised(input_df, output_df, cuda,
                input_tensoriser=input_tensoriser, output_tensoriser=output_tensoriser)

    def iter_batches(self, batch_size: int, shuffle: bool = False, input_only=False):
        yield from self._torchDataSet.iter_batches(batch_size, shuffle=shuffle, input_only=input_only)

    def size(self) -> Optional[int]:
        return self._torchDataSet.size()


class TorchDataSetProviderFromDataUtil(TorchDataSetProvider):
    def __init__(self, data_util: DataUtil, cuda: bool):
        super().__init__(input_tensor_scaler=data_util.get_input_tensor_scaler(), output_tensor_scaler=data_util.get_output_tensor_scaler(),
            input_dim=data_util.input_dim(), model_output_dim=data_util.model_output_dim())
        self.dataUtil = data_util
        self.cuda = cuda

    def provide_split(self, fractional_size_of_first_set: float) -> Tuple[TorchDataSet, TorchDataSet]:
        (x1, y1), (x2, y2) = self.dataUtil.split_into_tensors(fractional_size_of_first_set)
        return TorchDataSetFromTensors(x1, y1, self.cuda), TorchDataSetFromTensors(x2, y2, self.cuda)


class TorchDataSetProviderFromVectorDataUtil(TorchDataSetProvider):
    def __init__(self, data_util: VectorDataUtil, cuda: bool, tensorise_dynamically=False):
        super().__init__(input_tensor_scaler=data_util.get_input_tensor_scaler(), output_tensor_scaler=data_util.get_output_tensor_scaler(),
            input_dim=data_util.input_dim(), model_output_dim=data_util.model_output_dim())
        self.dataUtil = data_util
        self.cuda = cuda
        self.tensoriseDynamically = tensorise_dynamically

    def provide_split(self, fractional_size_of_first_set: float) -> Tuple[TorchDataSet, TorchDataSet]:
        return self.dataUtil.split_into_data_sets(fractional_size_of_first_set, self.cuda, tensorise_dynamically=self.tensoriseDynamically)


class TensorTransformer(ABC):
    @abstractmethod
    def transform(self, t: torch.Tensor) -> torch.Tensor:
        pass



import logging
from typing import Callable, Optional, Sequence, Union

import torch.nn.functional

from .mlp_modules import MultiLayerPerceptron
from ...torch_base import VectorTorchModel, TorchVectorRegressionModel, TorchVectorClassificationModel, ClassificationOutputMode
from ...torch_enums import ActivationFunction
from ...torch_opt import NNOptimiserParams
from .... import NormalisationMode

log: logging.Logger = logging.getLogger(__name__)


class MultiLayerPerceptronTorchModel(VectorTorchModel):
    def __init__(self,
            cuda: bool,
            hidden_dims: Sequence[int],
            hid_activation_function: Callable[[torch.Tensor], torch.Tensor],
            output_activation_function: Optional[Callable[[torch.Tensor], torch.Tensor]],
            p_dropout: Optional[float] = None,
            input_dim: Optional[int] = None) -> None:
        """
        :param cuda: whether to enable CUDA
        :param hidden_dims: the sequence of hidden layer dimensions
        :param hid_activation_function: the output activation function for hidden layers
        :param output_activation_function: the output activation function
        :param p_dropout: the dropout probability for training
        :param input_dim: the input dimension; if None, use dimensions determined by the input data (number of columns in data frame)
        """
        super().__init__(cuda=cuda)
        self.hidActivationFunction = ActivationFunction.torch_function_from_any(hid_activation_function)
        self.outputActivationFunction = ActivationFunction.torch_function_from_any(output_activation_function)
        self.hiddenDims = hidden_dims
        self.pDropout = p_dropout
        self.overrideInputDim = input_dim

    def create_torch_module_for_dims(self, input_dim: int, output_dim: int) -> torch.nn.Module:
        return MultiLayerPerceptron(input_dim if self.overrideInputDim is None else self.overrideInputDim, output_dim, self.hiddenDims,
            hid_activation_fn=self.hidActivationFunction, output_activation_fn=self.outputActivationFunction,
            p_dropout=self.pDropout)


class MultiLayerPerceptronVectorRegressionModel(TorchVectorRegressionModel):
    def __init__(self,
            hidden_dims: Sequence[int] = (5, 5),
            hid_activation_function: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            output_activation_function: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
            input_dim: Optional[int] = None,
            normalisation_mode: NormalisationMode = NormalisationMode.NONE,
            cuda: bool = True,
            p_dropout: Optional[float] = None,
            nn_optimiser_params: Optional[NNOptimiserParams] = None) -> None:
        """
        :param hidden_dims: sequence containing the number of neurons to use in hidden layers
        :param hid_activation_function: the activation function (torch.nn.functional.* or torch.*) to use for all hidden layers
        :param output_activation_function: the output activation function (torch.nn.functional.* or torch.* or None)
        :param input_dim: the input dimension; if None, use dimensions determined by the input data (number of columns in data frame)
        :param normalisation_mode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param p_dropout: the probability with which to apply dropouts after each hidden layer
        :param nn_optimiser_params: parameters for NNOptimiser; if None, use default (or what is specified in nnOptimiserDictParams)
        """
        self.hidden_dims = hidden_dims
        self.hid_activation_function = hid_activation_function
        self.output_activation_function = output_activation_function
        self.input_dim = input_dim
        self.cuda = cuda
        self.p_dropout = p_dropout
        super().__init__(self._create_torch_model, normalisation_mode, nn_optimiser_params)

    def _create_torch_model(self) -> MultiLayerPerceptronTorchModel:
        return MultiLayerPerceptronTorchModel(self.cuda, self.hidden_dims, self.hid_activation_function, self.output_activation_function,
            p_dropout=self.p_dropout, input_dim=self.input_dim)


class MultiLayerPerceptronVectorClassificationModel(TorchVectorClassificationModel):
    def __init__(self,
            hidden_dims: Sequence[int] = (5, 5),
            hid_activation_function: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            output_activation_function: Optional[Union[Callable[[torch.Tensor], torch.Tensor], str, ActivationFunction]] =
                ActivationFunction.LOG_SOFTMAX,
            input_dim: Optional[int] = None,
            normalisation_mode: NormalisationMode = NormalisationMode.NONE, cuda: bool = True,
            p_dropout: Optional[float] = None,
            nn_optimiser_params: Optional[NNOptimiserParams] = None) -> None:
        """
        :param hidden_dims: sequence containing the number of neurons to use in hidden layers
        :param hid_activation_function: the activation function (torch.nn.functional.* or torch.*) to use for all hidden layers
        :param output_activation_function: the output activation function (function from torch.nn.functional.*, function name, enum instance or None)
        :param input_dim: the input dimension; if None, use dimensions determined by the input data (number of columns in data frame)
        :param normalisation_mode: the normalisation mode to apply to input and output data
        :param cuda: whether to use CUDA (GPU acceleration)
        :param p_dropout: the probability with which to apply dropouts after each hidden layer
        :param nn_optimiser_params: parameters for NNOptimiser; if None, use default
        """
        self.hidden_dims = hidden_dims
        self.hid_activation_function = hid_activation_function
        self.output_activation_function = output_activation_function
        self.input_dim = input_dim
        self.cuda = cuda
        self.p_dropout = p_dropout
        output_mode = ClassificationOutputMode.for_activation_fn(ActivationFunction.torch_function_from_any(output_activation_function))
        super().__init__(output_mode,
            self._create_torch_model,
            normalisation_mode,
            nn_optimiser_params)

    def _create_torch_model(self) -> MultiLayerPerceptronTorchModel:
        return MultiLayerPerceptronTorchModel(self.cuda, self.hidden_dims, self.hid_activation_function, self.output_activation_function,
            p_dropout=self.p_dropout, input_dim=self.input_dim)


import logging
import os
import pickle
from pathlib import Path
from typing import List, Dict, Any, Iterable, Union

import joblib

from .io import is_s3_path, S3Object

log = logging.getLogger(__name__)


def load_pickle(path: Union[str, Path], backend="pickle"):
    if isinstance(path, Path):
        path = str(path)

    def read_file(f):
        if backend == "pickle":
            try:
                return pickle.load(f)
            except:
                log.error(f"Error loading {path}")
                raise
        elif backend == "joblib":
            return joblib.load(f)
        else:
            raise ValueError(f"Unknown backend '{backend}'")

    if is_s3_path(path):
        return read_file(S3Object(path).open_file("rb"))
    with open(path, "rb") as f:
        return read_file(f)


def dump_pickle(obj, pickle_path: Union[str, Path], backend="pickle", protocol=pickle.HIGHEST_PROTOCOL):
    if isinstance(pickle_path, Path):
        pickle_path = str(pickle_path)

    def open_file():
        if is_s3_path(pickle_path):
            return S3Object(pickle_path).open_file("wb")
        else:
            return open(pickle_path, "wb")

    dir_name = os.path.dirname(pickle_path)
    if dir_name != "":
        os.makedirs(dir_name, exist_ok=True)
    with open_file() as f:
        if backend == "pickle":
            try:
                pickle.dump(obj, f, protocol=protocol)
            except AttributeError as e:
                failing_paths = PickleFailureDebugger.debug_failure(obj)
                raise AttributeError(f"Cannot pickle paths {failing_paths} of {obj}: {str(e)}")
        elif backend == "joblib":
            joblib.dump(obj, f, protocol=protocol)
        else:
            raise ValueError(f"Unknown backend '{backend}'")


class PickleFailureDebugger:
    """
    A collection of methods for testing whether objects can be pickled and logging useful infos in case they cannot
    """

    enabled = False  # global flag controlling the behaviour of logFailureIfEnabled

    @classmethod
    def _debug_failure(cls, obj, path, failures, handled_object_ids):
        if id(obj) in handled_object_ids:
            return
        handled_object_ids.add(id(obj))

        try:
            pickle.dumps(obj)
        except:
            # determine dictionary of children to investigate (if any)
            if hasattr(obj, '__dict__'):  # Because of strange behaviour of getstate, here try-except is used instead of if-else
                try:  # Because of strange behaviour of getattr(_, '__getstate__'), we here use try-except
                    d = obj.__getstate__()
                    if type(d) != dict:
                        d = {"state": d}
                except:
                    d = obj.__dict__
            elif type(obj) == dict:
                d = obj
            elif type(obj) in (list, tuple, set):
                d = dict(enumerate(obj))
            else:
                d = {}

            # recursively test children
            have_failed_child = False
            for key, child in d.items():
                child_path = list(path) + [f"{key}[{child.__class__.__name__}]"]
                have_failed_child = cls._debug_failure(child, child_path, failures, handled_object_ids) or have_failed_child

            if not have_failed_child:
                failures.append(path)

            return True
        else:
            return False

    @classmethod
    def debug_failure(cls, obj) -> List[str]:
        """
        Recursively tries to pickle the given object and returns a list of failed paths

        :param obj: the object for which to recursively test pickling
        :return: a list of object paths that failed to pickle
        """
        handled_object_ids = set()
        failures = []
        cls._debug_failure(obj, [obj.__class__.__name__], failures, handled_object_ids)
        return [".".join(l) for l in failures]

    @classmethod
    def log_failure_if_enabled(cls, obj, context_info: str = None):
        """
        If the class flag 'enabled' is set to true, the pickling of the given object is
        recursively tested and the results are logged at error level if there are problems and
        info level otherwise.
        If the flag is disabled, no action is taken.

        :param obj: the object for which to recursively test pickling
        :param context_info: optional additional string to be included in the log message
        """
        if cls.enabled:
            failures = cls.debug_failure(obj)
            prefix = f"Picklability analysis for {obj}"
            if context_info is not None:
                prefix += " (context: %s)" % context_info
            if len(failures) > 0:
                log.error(f"{prefix}: pickling would result in failures due to: {failures}")
            else:
                log.info(f"{prefix}: is picklable")


def setstate(cls,
        obj,
        state: Dict[str, Any],
        renamed_properties: Dict[str, str] = None,
        new_optional_properties: List[str] = None,
        new_default_properties: Dict[str, Any] = None,
        removed_properties: List[str] = None) -> None:
    """
    Helper function for safe implementations of __setstate__ in classes, which appropriately handles the cases where
    a parent class already implements __setstate__ and where it does not. Call this function whenever you would actually
    like to call the super-class' implementation.
    Unfortunately, __setstate__ is not implemented in object, rendering super().__setstate__(state) invalid in the general case.

    :param cls: the class in which you are implementing __setstate__
    :param obj: the instance of cls
    :param state: the state dictionary
    :param renamed_properties: a mapping from old property names to new property names
    :param new_optional_properties: a list of names of new property names, which, if not present, shall be initialised with None
    :param new_default_properties: a dictionary mapping property names to their default values, which shall be added if they are not present
    :param removed_properties: a list of names of properties that are no longer being used
    """
    # handle new/changed properties
    if renamed_properties is not None:
        for mOld, mNew in renamed_properties.items():
            if mOld in state:
                state[mNew] = state[mOld]
                del state[mOld]
    if new_optional_properties is not None:
        for mNew in new_optional_properties:
            if mNew not in state:
                state[mNew] = None
    if new_default_properties is not None:
        for mNew, mValue in new_default_properties.items():
            if mNew not in state:
                state[mNew] = mValue
    if removed_properties is not None:
        for p in removed_properties:
            if p in state:
                del state[p]
    # call super implementation, if any
    s = super(cls, obj)
    if hasattr(s, '__setstate__'):
        s.__setstate__(state)
    else:
        obj.__dict__ = state


def getstate(cls,
        obj,
        transient_properties: Iterable[str] = None,
        excluded_properties: Iterable[str] = None,
        override_properties: Dict[str, Any] = None,
        excluded_default_properties: Dict[str, Any] = None) -> Dict[str, Any]:
    """
    Helper function for safe implementations of __getstate__ in classes, which appropriately handles the cases where
    a parent class already implements __getstate__ and where it does not. Call this function whenever you would actually
    like to call the super-class' implementation.
    Unfortunately, __getstate__ is not implemented in object, rendering super().__getstate__() invalid in the general case.

    :param cls: the class in which you are implementing __getstate__
    :param obj: the instance of cls
    :param transient_properties: transient properties which be set to None in serialisations
    :param excluded_properties: properties which shall be completely removed from serialisations
    :param override_properties: a mapping from property names to values specifying (new or existing) properties which are to be set;
        use this to set a fixed value for an existing property or to add a completely new property
    :param excluded_default_properties: properties which shall be completely removed from serialisations, if they are set
        to the given default value
    :return: the state dictionary, which may be modified by the receiver
    """
    s = super(cls, obj)
    if hasattr(s, '__getstate__'):
        d = s.__getstate__()
    else:
        d = obj.__dict__.copy()
    if transient_properties is not None:
        for p in transient_properties:
            if p in d:
                d[p] = None
    if excluded_properties is not None:
        for p in excluded_properties:
            if p in d:
                del d[p]
    if override_properties is not None:
        for k, v in override_properties.items():
            d[k] = v
    if excluded_default_properties is not None:
        for p, v in excluded_default_properties.items():
            if p in d and d[p] == v:
                del d[p]
    return d
